import torch


def exp2r( w ):
    """ Convert R^3 tensor to SO(3) tensor

    Args:
        w - (N, 3) angular rotation vector

    Return:
        R - (N, 3, 3) Rotation matrices


    """
    # compute norms and normalized omega vectors
    thetas = torch.norm( w, p='fro', dim=1, keepdim=True )
    norms = thetas * (thetas != 0) + torch.ones_like( thetas ) * (thetas == 0)
    w_norm = w / norms
    W_norm = skew( w_norm )

    Rmat = (
            torch.eye( 3, dtype=w.dtype, device=w.device )[ None ].repeat( w.shape[ 0 ], 1, 1 )
            + torch.sin( thetas.unsqueeze( dim=-1 ) ) * W_norm
            + (1 - torch.cos( thetas.unsqueeze( dim=-1 ) )) * (W_norm @ W_norm)
    )

    return Rmat


# exp2r

def skew( w ):
    """ Skewify R^3 tensor

    Args:
        w - (N, 3)

    Return:
        W - (N, 3, 3) skew-symmetric matrix

    """
    zeros = torch.zeros_like( w[ :, 0:1 ], dtype=w.dtype, device=w.device )
    W_half = torch.cat(
            (
                    zeros, -w[ :, 2:3 ], w[ :, 1:2 ],
                    zeros, zeros, -w[ :, 0:1 ],
                    zeros, zeros, zeros,
            ),
            dim=1
    ).reshape( -1, 3, 3 )

    return W_half - W_half.transpose( 1, 2 )

# skew



import tensorflow as tf

from needle_shape_sensing.tensorflow import geometry


def integrateEP_w0(
        w_init: tf.Tensor, w0: tf.Tensor, w0prime: tf.Tensor, B: tf.Tensor,
        seq_mask: tf.Tensor, ds: float, R_init: tf.Tensor = tf.eye( 3 ),
        Binv: tf.Tensor = None, wv_only: bool = False
):
    """
    integrate Euler-Poincare equation for needle shape sensing for given intrinsic angular deformation

    Args:
            w_init: (N, 3) initial deformation vector
            w0: (N, M, 3) intrinsic angular deformation
            w0prime: (N, M, 3) d/ds w0
            B: 3 x 3 needle stiffness matrix
            seq_mask: (N, M) mask of which are valid points to use
            ds: the arclength increments
            Binv: (Default = None) inv(B) Can be provided for numerical efficiency
            R_init: (Default = 3x3 identity) SO3 matrix for initial rotation angle
            wv_only: (Default = False) whether to only integrate wv or not.

        Return:
            (N, M, 3) needle shape, (N, M, 3, 3) SO3 matrices of orientations), (N, M, 3) angular deformation), (N, M) sequence mask
            (None, None, (N, M, 3) angular deformation) if 'wv_only' is True

    """

    if Binv is None:
        Binv = tf.linalg.inv( B )

    # if

    # unpack the shapes
    N, M = w0.shape[ 0:2 ]

    # tensor-ify B and Binv
    B_T = tf.cast(
            tf.tile(
                    tf.reshape( B, (1, B.shape[ 0 ], B.shape[ 1 ]) ),
                    (N, 1, 1)
            ), dtype=w0.dtype
    )
    Binv_T = tf.cast(
            tf.tile(
                    tf.reshape( Binv, (1, B.shape[ 0 ], B.shape[ 1 ]) ),
                    (N, 1, 1)
            ), dtype=w0.dtype
    )

    # prepare tensors
    wv = tf.TensorArray(
            w0.dtype,
            size=M,
            dynamic_size=False,
            infer_shape=True,
            clear_after_read=False,
    )
    wv = wv.write( 0, w_init )

    for idx in range( 1, M ):
        seq_mask_idx = tf.expand_dims( seq_mask[ :, idx:idx + 1 ], axis=-1 )

        # unpack veriables
        w0_im1      = tf.expand_dims( w0[ :, idx - 1 ], axis=-1 )
        w0prime_im1 = tf.expand_dims( w0prime[ :, idx - 1 ], axis=-1 )
        wv_im1      = tf.expand_dims( wv.read( idx - 1 ), axis=-1 )

        if idx == 1:
                wv_i = (
                        wv_im1 + 1 * ds * (
                                w0prime_im1 - Binv_T @ tf.linalg.cross(
                                        tf.squeeze(wv_im1, axis=-1),
                                        tf.squeeze(B_T @ (wv_im1 - w0_im1), axis=-1)
                                 )[:, :, tf.newaxis]
                        )
                ) * tf.cast( seq_mask_idx, dtype=wv.dtype )

        # if
        else:
            wv_im2 = tf.expand_dims( wv.read( idx - 2 ), axis=-1 )
            wv_i   = (
                        wv_im2 + 2 * ds * (
                                w0prime_im1 - Binv_T @ tf.linalg.cross(
                                        tf.squeeze(wv_im1, axis=-1),
                                        tf.squeeze(B_T @ (wv_im1 - w0_im1), axis=-1)
                                )[:, :, tf.newaxis]
                        )
                ) * tf.cast( seq_mask_idx, dtype=wv.dtype )

        wv = wv.write( idx, tf.squeeze( wv_i, axis=-1 ) )

    # for

    wv = tf.transpose( wv.stack(), [ 1, 0, 2 ] )

    if wv_only:
        pmat, Rmat = None, None

    else:
        pmat, Rmat, seq_mask = integratePose_wv( wv, seq_mask, ds=ds, R_init=R_init, )

    return pmat, Rmat, wv, seq_mask


# integrateEP_w0

def integratePose_wv(
        wv: tf.Tensor, seq_mask: tf.Tensor, ds: float, R_init=tf.eye( 3 )
):
    """ Integrate angular deformation to get the pose of the needle along it's arclengths

        :param wv: (N, M, 3) angular deformation vector
        :param seq_mask: (N, M) sequence mask for which are viable
        :param ds: (Default = None) the arclength increments desired
        :param R_init: (Default = numpy.eye(3)) Rotation matrix of the inital pose

        :returns: pmat, Rmat, sequence mask
            - pmat: (N, M, 3) position for the needle shape points in-tissue
            - Rmat: (N, M, 3, 3) SO(3) rotation matrices for
            - sequence mask: (N, M) of boolean for which sequences are viable
    """
    N, M = wv.shape[ 0:2 ]

    pmat_ta = tf.TensorArray(
            wv.dtype,
            size=M,
            dynamic_size=False,
            infer_shape=True,
            clear_after_read=False,
    )  # (M, N, 3)
    Rmat_ta = tf.TensorArray(
            wv.dtype,
            size=M,
            dynamic_size=False,
            infer_shape=True,
            clear_after_read=False,
    )  # (M, N, 3, 3)

    # initial conditions
    pmat_ta = pmat_ta.write( 0, tf.zeros( (N, 3), dtype=pmat_ta.dtype ) )
    Rmat_ta = Rmat_ta.write(
            0, tf.cast(
                    tf.tile(
                            tf.reshape( R_init, (1, R_init.shape[ 0 ], R_init.shape[ 1 ]) ),
                            (N, 1, 1)
                    ), dtype=Rmat_ta.dtype
            )
    )

    for idx in range( 1, M ):
        # unpack vars
        seq_mask_i = seq_mask[ :, idx ]
        Rmat_im1 = Rmat_ta.read( idx - 1 )  # (N, 3, 3)
        pmat_im1 = pmat_ta.read( idx - 1 )  # (N, 3)

        # integrate
        Rmat_i = Rmat_im1 @ (
                geometry.exp2r(
                        ds * tf.reduce_mean( wv[ :, idx - 1:idx ], axis=1 )
                ) * tf.cast( seq_mask_i[ :, tf.newaxis, tf.newaxis ], dtype=wv.dtype )
        )

        # add results for Rmat
        Rmat_ta = Rmat_ta.write( idx, Rmat_i )

        e3vec = tf.transpose( Rmat_ta.stack()[ :idx + 1, :, :, 2 ], [ 1, 0, 2 ] )  # grab z-directions
        if idx == 1:
            pmat_i = pmat_im1 + Rmat_i[ :, :, 2 ] * ds

        else:
            pmat_i = simpson_vec_int( e3vec, ds, seq_mask[ :, :idx + 1 ] )

        # add results for pmat
        pmat_ta = pmat_ta.write( idx, pmat_i )

    # for

    Rmat = tf.transpose( Rmat_ta.stack(), [ 1, 0, 2, 3 ] )
    pmat = tf.transpose( pmat_ta.stack(), [ 1, 0, 2 ] )

    return pmat, Rmat, seq_mask


# integratePose_wv


def simpson_vec_int( f: tf.Tensor, dx: float, seq_mask: tf.Tensor ):
    """ Implementation of Simpson vector integration for tensor integration

            Original Author (MATLAB): Jin Seob Kim
            Translated Author: Dimitri Lezcano

            Args:
                 f:  (N, M, D) tensor where D is the dimension of the vector is the dimension of the parameter ( N > 2 )
                        and M Integration samples
                 dx: float of the step size
                 seq_mask: (N, M) boolean mask for which parts of the integration is valid

            Return:
                integrated tensor of shape (N, D)

    """
    N, M, D = f.shape[ :3 ]
    num_intervals = M - 1
    assert (num_intervals > 1)

    f_masked = f * tf.cast( seq_mask[ :, :, tf.newaxis ], dtype=f.dtype )

    # perform the integration
    int_res = tf.zeros( (f.shape[ 0 ], f.shape[ 2 ]), dtype=f.dtype )
    if num_intervals == 2:
        int_res = dx / 3 * tf.reduce_sum(
                f_masked[ :, 0:3 ] * tf.cast( tf.reshape( [ 1., 4., 1. ], (1, -1, 1) ), dtype=f_masked.dtype ),
                axis=1
        )

        return int_res

    # if

    elif num_intervals == 3:
        int_res = 3 / 8 * dx * tf.reduce_sum(
                f_masked[ :, 0:4 ] * tf.cast( tf.reshape( [ 1., 3., 3., 1. ], (1, -1, 1) ), dtype=f_masked.dtype ),
                axis=1
        )

        return int_res

    # elif

    if num_intervals % 2 != 0:
        int_res = (
                int_res +
                3 / 8 * dx * tf.reduce_sum(
                f_masked[ :, -4: ] * tf.cast( tf.reshape( [ 1., 3., 3., 1. ], (1, -1, 1) ), dtype=f_masked.dtype ),
                axis=1
        )
        )
        m = num_intervals - 3

    # if
    else:
        m = num_intervals

    # else

    int_res = (
            int_res
            + dx / 3 * (
                    f_masked[ :, 0 ] + f_masked[ :, m ]
                    + 4 * tf.reduce_sum(
                    f_masked[ :, 1:m:2 ], axis=1
            )
            )
    )

    if m > 2:
        int_res = (
                int_res
                + 2 / 3 * dx * tf.reduce_sum(
                f_masked[ :, 2:m:2 ], axis=1
        )
        )
    # if

    return int_res

# simpson_vec_int



import numpy as np

from needle_shape_sensing import numerical, intrinsics, geometry, sensorized_needles


class ShapeSensingFBGNeedle( sensorized_needles.FBGNeedle ):
    def __init__(
            self, length: float, serial_number: str, num_channels: int, sensor_location=None,
            calibration_mats=None, weights=None, ds: float = 0.5, current_depth: float = 0,
            optim_options: dict = None, cts_integration: bool = False, **kwargs ):
        super().__init__(
                length,
                serial_number,
                num_channels,
                sensor_location=sensor_location,
                calibration_mats=calibration_mats,
                weights=weights,
                **kwargs
        )

        # current insertion parameters
        self.current_depth        = current_depth
        self.__current_shapetype  = intrinsics.SHAPETYPE.SINGLEBEND_SINGLELAYER
        self.current_wavelengths  = -np.ones_like( self.ref_wavelengths )
        self.current_curvatures   = np.zeros( (2, self.num_activeAreas), dtype=self.ref_wavelengths.dtype )
        self.insertion_point      = np.zeros((3,), dtype=self.ref_wavelengths.dtype) # relative to needle
        self.insertion_parameters = dict()

        # define needle shape-sensing optimizers
        self.optimizer = numerical.NeedleParamOptimizations(
            self,
            ds=ds,
            optim_options=optim_options,
            continuous=cts_integration
        )
        self.current_kc = [ 0 ]
        self.current_rotations = None # [ 0 ] * int(self.length // self.ds + 1)  # radians
        self.current_winit = np.zeros( 3 )

    # __init__

    def __repr__( self ):
        return "Shape Sensing " + super().__repr__()

    @property
    def current_shapetype( self ):
        return self.__current_shapetype

    # property: current_shapetype

    @property
    def ds( self ):
        return self.optimizer.ds

    # property: ds

    @ds.setter
    def ds( self, ds ):
        self.optimizer.ds = ds

    # property setter: ds

    @property
    def continuous_integration( self ):
        return self.optimizer.continuous

    # continuous_integration

    @continuous_integration.setter
    def continuous_integration( self, continuous: bool ):
        self.optimizer.continuous = continuous

    # property setter: continuous_integration

    @staticmethod
    def from_FBGNeedle( fbgneedle: sensorized_needles.FBGNeedle, **kwargs ):
        """ Turn an FBGNeedle into a shape-sensing FBGNeedle

            :param fbgneedle: FBGNeedle to turn into a sensorized one
            :keyword ds: the ds for the ShapeSensingFBGNeedle constructor
            :keyword current_depth: the current insertion depth for the ShapeSensingFBGNeedle constructor

            :return: ShapeSensingFBGNeedle with the current FBGNeedle
        """
        if isinstance(fbgneedle, sensorized_needles.MCFNeedle):
            return ShapeSensingMCFNeedle.from_MCFNeedle(fbgneedle, **kwargs)

        return ShapeSensingFBGNeedle(
                fbgneedle.length,
                fbgneedle.serial_number,
                fbgneedle.num_channels,
                sensor_location=fbgneedle.sensor_location,
                calibration_mats=fbgneedle.cal_matrices,
                weights=fbgneedle.weights,
                diameter=fbgneedle.diameter,
                Emod=fbgneedle.Emod,
                pratio=fbgneedle.pratio,
                ref_wavelengths=fbgneedle.ref_wavelengths,
                **kwargs
        )

    # from_FBGNeedle

    def get_needle_shape( self, *args, **kwargs ):
        """ Determine the 3D needle shape of the current shape-sensing needle within a specific insertion depth

        Example (Single-Bend Single-Layer)
            pmat, Rmat = ss_fbgneedle.get_needle_shape(kc_i, w_init_i, R_init=R_init)

        Example (Single-Bend Double-Layer)
            pmat, Rmat = ss_fbgneedle.get_needle_shape(kc_i, kc2_i, w_init_i, R_init=R_init)

        Example (Double-Bend Single-Layer)
            pmat, Rmat = ss_fbgneedle.get_needle_shape(kc_i, w_init_i, R_init=R_init)

        :param kcx_i: the initial kappa_c value(s)
        :param w_init_i: (Default = None) the omega_init value
        :keyword R_init: (Default = numpy.eye(3)) the initial angular offset
        :keyword air_cubic_fit: (Default = False) whether to handle air-gap with
        :keyword return_wv: (Default = False) whether to return wv or not
        :returns: (N x 3 position matrix of the needle shape, N x 3 x 3 orientation matrices of the needle shape, N x 3 curvature matrix (if return_wv=True))
                  (None, None, None (if return_wv=True)) if:
                        - sensors are not calibrated
                        - current insertion depth not > 0
                        - current shape type is not an implemented shape type
        """
        # kwargs get
        R_init = kwargs.get( 'R_init', np.eye( 3 ) )
        s = np.arange( 0, self.current_depth + self.ds, self.ds )
        current_rotations = None if self.current_rotations is None else self.current_rotations[ -int(self.current_depth // self.ds - 1): ]


        # initial checks
        pmat, Rmat, wv = None, None, None
        if not self.sensor_calibrated:  # check current wavelengths
            return pmat, Rmat

        if self.current_shapetype not in intrinsics.SHAPETYPE:
            return pmat, Rmat

        # handle air-gap
        pmat_air, Rmat_air, wv_air = None, None, None
        if self.insertion_point[2] > self.ds:
            s_air = np.arange(
                0,
                (self.insertion_point[2]//self.ds + 1) * self.ds,
                self.ds
            )
            pmat_air, Rmat_air = intrinsics.AirDeflection.shape(
                s_air,
                self.insertion_point,
                cubic_fit=kwargs.get("air_cubic_fit", False),
            )
            # Rmat_air = np.tile(np.eye(3), (pmat_air.shape[0], 1, 1)) # FIXME: update with model-based method
            wv_air   = np.zeros_like(pmat_air) # FIXME: update with model-based method

        # if

        if self.current_depth <= 0:  # check insertion depth
            return pmat_air, Rmat_air

        # initalization
        k0, k0prime, w_init = None, None, None

        if self.current_shapetype == intrinsics.SHAPETYPE.CONSTANT_CURVATURE:
            # determine parameters
            curvature = self.optimizer.constant_curvature(
                    self.current_curvatures.T, self.current_depth )
            curvature = np.append( curvature, 0 )  # ensure 3 vector
            pmat, Rmat = intrinsics.ConstantCurvature.shape( s, curvature )

            pmat = pmat @ R_init.T
            Rmat = R_init @ Rmat
            wv   = np.tile(curvature, (pmat.shape[0], 1))

            self.current_winit = curvature
            self.current_kc    = [np.linalg.norm(curvature)]

        # if
        elif self.current_shapetype == intrinsics.SHAPETYPE.SINGLEBEND_SINGLELAYER:
            # get parameters
            kc_i = args[ 0 ]
            if len( args ) > 1:
                w_init_i = args[ 1 ]
                if not isinstance( w_init_i, np.ndarray ) or len( w_init_i ) != 3:
                    raise ValueError( "w_init_i must be a 3-D vector" )
            else:
                w_init_i = np.array( [ kc_i, 0, 0 ] )

            # else

            # determine parameters
            kc, w_init, _ = self.optimizer.singlebend_singlelayer_k0(
                kc_i,
                w_init_i,
                self.current_curvatures.T,
                self.current_depth,
                R_init=R_init,
                needle_rotations=current_rotations,
                **kwargs
            )
            self.current_kc = [ kc ]
            self.current_winit = w_init

            # determine k0 and k0prime
            k0, k0prime = intrinsics.SingleBend.k0_1layer(
                s,
                kc,
                self.current_depth,
                return_callable=self.continuous_integration
            )

        # if: single-bend single-layer

        elif self.current_shapetype == intrinsics.SHAPETYPE.SINGLEBEND_DOUBLELAYER:
            # get parameters
            z_crit = self.insertion_parameters[ 'z_crit' ]
            kc1_i = args[ 0 ]
            kc2_i = args[ 1 ]
            if len( args ) > 2:
                w_init_i = args[ 2 ]
                if not isinstance( w_init_i, np.ndarray ) or len( w_init_i ) != 3:
                    raise ValueError( "w_init_i must be a 3-D vector" )
            else:
                w_init_i = np.array( [ kc1_i, 0, 0 ] )

            # else

            # determine parameters
            kc1, kc2, w_init, _ = self.optimizer.singlebend_doublelayer_k0(
                kc1_i,
                kc2_i,
                w_init_i,
                self.current_curvatures.T,
                self.current_depth,
                z_crit=z_crit,
                R_init=R_init,
                needle_rotations=current_rotations,
            )
            self.current_kc = [ kc1, kc2 ]
            self.current_winit = w_init
            s_crit = intrinsics.SingleBend.determine_2layer_boundary(
                kc1,
                self.current_depth,
                z_crit,
                self.B,
                w_init=w_init,
                s0=0,
                ds=self.ds,
                R_init=R_init,
                needle_rotations=current_rotations,
                continuous=self.continuous_integration,
            )

            # determine k0 and k0prime
            k0, k0prime = intrinsics.SingleBend.k0_2layer(
                s,
                kc1,
                kc2,
                self.current_depth,
                s_crit,
                return_callable=self.continuous_integration,
            )
            Rz     = geometry.rotz( np.pi )
            R_init = R_init @ Rz  # rotate the needle 180 degrees about its axis

        # elif: single-bend double-layer

        elif self.current_shapetype == intrinsics.SHAPETYPE.DOUBLEBEND_SINGLELAYER:
            # get parameters
            s_crit = self.insertion_parameters[ 's_double_bend' ]
            kc_i = args[ 0 ]
            if len( args ) > 1:
                w_init_i = args[ 1 ]
                if not isinstance( w_init_i, np.ndarray ) or len( w_init_i ) != 3:
                    raise ValueError( "w_init_i must be a 3-D vector" )
            else:
                w_init_i = np.array( [ kc_i, 0, 0 ] )

            # else

            kc, w_init, _ = self.optimizer.doublebend_singlelayer_k0(
                kc_i,
                w_init_i,
                self.current_curvatures.T,
                self.current_depth,
                s_crit,
                R_init=R_init,
            )
            self.current_kc = [ kc ]
            self.current_winit = w_init
            k0, k0prime = intrinsics.DoubleBend.k0_1layer(
                s,
                kc,
                self.current_depth,
                s_crit=s_crit,
                return_callable=self.continuous_integration,
            )

        # elif: double-bend single-layer

        else:
            return pmat, Rmat

        # else: Cannot find parameterization


        # compute w0, w0prime and shape
        if self.continuous_integration:
            w0      = lambda s: np.append( k0( s ), [ 0, 0 ] )
            w0prime = lambda s: np.append( k0prime( s ), [ 0, 0 ] )

            pmat, Rmat, wv = numerical.integrateEP_w0_ode(
                w_init,
                w0,
                w0prime,
                self.B,
                s0=0,
                ds=self.ds,
                needle_rotations=self.current_rotations,
                R_init=R_init,
                arg_check=False,
            )
        # if
        else:
            w0      = np.hstack( (k0.reshape( -1, 1 ), np.zeros( (k0.size, 2) )) )
            w0prime = np.hstack( (k0prime.reshape( -1, 1 ), np.zeros( (k0prime.size, 2) )) )

            pmat, Rmat, wv = numerical.integrateEP_w0(
                w_init,
                w0,
                w0prime,
                self.B,
                s0=0,
                ds=self.ds,
                needle_rotations=self.current_rotations,
                R_init=R_init,
                arg_check=False,
            )
        # if

        # stack the air layer
        if (pmat_air is not None) and (Rmat_air is not None) and (wv_air is not None):
            pmat = pmat @ Rmat_air[-1].T  + pmat_air[np.newaxis, -1]
            Rmat = Rmat_air[np.newaxis, -1] @ Rmat

            pmat = np.concatenate((pmat_air, pmat[1:]), axis=0)
            Rmat = np.concatenate((Rmat_air, Rmat[1:]), axis=0)
            wv   = np.concatenate((wv_air,   wv[1:]),   axis=0)

        # if

        elif self.insertion_point[2] > 0:
            pmat += self.insertion_point[np.newaxis]

        # elif

        return pmat, Rmat

    # get_needle_shape

    @staticmethod  # overloaded
    def load_json( filename: str ):
        """ Load a ShapeSensingFBGNeedle from a needle parameter json file"""
        return ShapeSensingFBGNeedle.from_FBGNeedle(
                super( ShapeSensingFBGNeedle, ShapeSensingFBGNeedle ).load_json( filename ) )

    # load_json

    def set_rotation( self, L: float, rot_rads: float ):
        """ Set the needle rotation at the specified length

            :param L: float of needle depth when the needle the rotation was performed
            :param rot_rads: float of the amount of rotation/orientation of the needle

            :returns: boolean of whether operation was successful or not

        """
        # check for valid length
        if (L < 0) or L > self.length:
            return False

        # find closest length
        lengths = np.arange( 0, self.length + self.ds, self.ds )
        L_closest_idx = np.argmin( np.abs( lengths - L ) ).item()

        # update the rotation # TODO: is this correct?
        # for i in range(L_closest_idx, len(self.current_rotations)): # update after points
        for i in range(L_closest_idx): # update before points
            self.current_rotations[i] = rot_rads

        return True

    # set_rotation

    def update_curvatures( self, processed: bool = False, temp_comp: bool = True ):
        """ Update the current curvatures

            :param processed: (Default = False) a boolean on whether the current signals are processed or not
            :param temp_comp: (Default = True) a boolean on whether to perform temperature compensated for non-processed signals

            :returns: True if update was successful, False otherwise.

        """
        if self.sensor_calibrated and self.is_calibrated:
            success = True

            if processed:
                self.current_curvatures = 1e-3 * self.curvatures_processed( self.current_wavelengths )

            else:
                self.current_curvatures = 1e-3 * self.curvatures_raw(
                        self.current_wavelengths, temp_comp=temp_comp )

        # if
        else:
            success = False

        # else

        return success

    # update_curvatures

    def update_shapetype( self, shapetype, *args ):
        """ Update the current needle shape-sensing type

            Example:
                ss_fbgneedle.update_shapetype(intrinsics.SHAPETYPES.SINGLEBEND_SINGLELAYER)

            Example:
                ss_fbgneedle.update_shapetype(intrinsics.SHAPETYPES.SINGLEBEND_DOUBLELAYER) -> throws IndexOutOfBoundsError
                ss_fbgneedle.update_shapetype(intrinsics.SHAPETYPES.SINGLEBEND_DOUBLELAYER, z_crit) -> OK

            :param shapetype: listing from current implemented shape-types
                if shapetype is SINGLEBEND_DOUBLELAYER:
                    first argument must be length of the first layer

                if shapetype is DOUBLEBENG_SINGLELAYER:
                    first argument must the double bend insertion depth


            :returns: True if update was successful, False otherwise.

        """
        if shapetype == intrinsics.SHAPETYPE.CONSTANT_CURVATURE:
            self.__current_shapetype = shapetype
            self.insertion_parameters = { }
            success = True

        # if
        elif shapetype == intrinsics.SHAPETYPE.SINGLEBEND_SINGLELAYER:
            self.__current_shapetype = shapetype
            self.insertion_parameters = { }
            success = True

        # elif
        elif shapetype == intrinsics.SHAPETYPE.SINGLEBEND_DOUBLELAYER:
            self.__current_shapetype = shapetype
            self.insertion_parameters = { 'z_crit': args[ 0 ] }
            success = True

        # elif
        elif shapetype == intrinsics.SHAPETYPE.DOUBLEBEND_SINGLELAYER:
            self.__current_shapetype = shapetype
            self.insertion_parameters = { 's_double_bend': args[ 0 ] }
            success = True

        # elif
        else:
            success = False

        # else

        return success

    # update_shapetype

    def update_wavelengths(
            self, wavelengths: np.ndarray, reference: bool = False, temp_comp: bool = True,
            processed: bool = False ):
        """ Update the current signals and curvatures with the updated value. This will also determine the
            curvatures if the current reference signals are set

            :param wavelengths: numpy array to update the current signals with
            :param reference: (Default = False) whether to update the reference wavelengths
            :param temp_comp: (Default = True) whether to perform temperature compensation on the signals
            :param processed: (Default = False) whether the signals are processed or not

        """
        # get mean value if it is an array of measurements
        if wavelengths.ndim == 2:
            wavelengths = np.mean( wavelengths, axis=0 )

        # if

        # check for wavlength shape to mach reference wavelengths shape
        if wavelengths.shape != self.ref_wavelengths.shape:
            raise IndexError( f"Signals must be of size {self.ref_wavelengths.shape}." )

        # if

        # set the wavelengths
        if reference:
            self.ref_wavelengths = wavelengths
            self.current_wavelengths = wavelengths

        # if
        else:
            self.current_wavelengths = wavelengths

        # else

        # calculate the curvatures
        success    = self.update_curvatures( processed=processed, temp_comp=temp_comp )
        curvatures = self.current_curvatures if success else None

        return wavelengths, curvatures  # return the signals & curvatures anyways

    # update_wavelengths

# class: ShapeSensingFBGNeedle

class ShapeSensingMCFNeedle(ShapeSensingFBGNeedle, sensorized_needles.MCFNeedle):
    def __init__(
        self,
        length:
        float,
        serial_number: str,
        num_channels: int,
        sensor_location=None,
        calibration_mats=None,
        central_core_ch:int = None,
        weights=None,
        ds: float = 0.5,
        current_depth: float = 0,
        optim_options: dict = None,
        cts_integration: bool = False,
        **kwargs
    ):
        super().__init__(
            length=length,
            serial_number=serial_number,
            num_channels=num_channels,
            sensor_location=sensor_location,
            calibration_mats=calibration_mats,
            weights=weights,
            ds=ds,
            current_depth=current_depth,
            optim_options=optim_options,
            cts_integration=cts_integration,
            central_core_ch=central_core_ch,
            **kwargs
        )

    # __init__

    @staticmethod
    def from_FBGNeedle( fbgneedle: sensorized_needles.FBGNeedle, **kwargs ):
        raise NotImplementedError("use 'from_MCFNeedle' function")

    # from_FBGNeedle

    @staticmethod
    def from_MCFNeedle( mcfneedle: sensorized_needles.MCFNeedle, **kwargs ):
        """ Turn an FBGNeedle into a shape-sensing FBGNeedle

            :param mcfneedle: MCFNeedle to turn into a sensorized one
            :keyword ds: the ds for the ShapeSensingMCFNeedle constructor
            :keyword current_depth: the current insertion depth for the ShapeSensingMCFNeedle constructor

            :return: ShapeSensingMCFNeedle with the current MCFNeedle
        """
        return ShapeSensingMCFNeedle(
                mcfneedle.length,
                mcfneedle.serial_number,
                mcfneedle.num_channels,
                sensor_location=mcfneedle.sensor_location,
                calibration_mats=mcfneedle.cal_matrices,
                weights=mcfneedle.weights,
                diameter=mcfneedle.diameter,
                Emod=mcfneedle.Emod,
                pratio=mcfneedle.pratio,
                ref_wavelengths=mcfneedle.ref_wavelengths,
                central_core_ch=mcfneedle.central_core_ch,
                **kwargs
        )

    # from_MCFNeedle

# ShapeSensingMCFNeedle


import torch


def get_device( acceleration: bool = False ):
    """ Gets the device to put the tensors on

    Args:
        acceleartion (bool, Default=False): whether to return a device (if available for acceleration)

    Returns:
        torch.device: the torch device to use

    """
    device = torch.device( "cpu" )

    if not acceleration:
        return device

    if torch.cuda.is_available():
        device = torch.device( "cuda" )

    elif torch.backends.mps.is_built():
        device = torch.device( "mps" )

    return device

# get_device



import tensorflow as tf


def exp2r( w ):
    """ Convert R^3 tensor to SO(3) tensor

    Args:
        w - (N, 3) angular rotation vector

    Return:
        R - (N, 3, 3) Rotation matrices


    """
    Rta = tf.TensorArray(
            w.dtype,
            size=w.shape[ 0 ],
            clear_after_read=False,
            dynamic_size=False,
    )

    W = skew( w )

    for i in range( w.shape[ 0 ] ):
        theta = tf.norm( w[ i ], ord='euclidean' )
        if theta == 0:
            Rta.write( i, tf.eye( 3, dtype=w.dtype ) )
            continue

        # iF

        R_i = (
                tf.eye( 3, dtype=Rta.dtype )
                + tf.sin( theta ) * W[ i ] / theta
                + (1 - tf.cos( theta )) * (W[ i ] @ W[ i ]) / theta ** 2
        )

        Rta.write( i, R_i )

    # for

    Rmat = Rta.stack()

    return Rmat


# exp2r

def skew( w ):
    """ Skewify R^3 tensor

    Args:
        w - (N, 3)

    Return:
        W - (N, 3, 3) skew-symmetric matrix

    """

    W_ta = tf.TensorArray(
            w.dtype,
            size=w.shape[ 0 ],
            clear_after_read=False,
            dynamic_size=False,
    )

    for i in range( w.shape[ 0 ] ):
        W_ta.write(
                i,
                [
                        [ 0, -w[ i, 2 ], w[ i, 1 ] ],
                        [ w[ i, 2 ], 0, -w[ i, 0 ] ],
                        [ -w[ i, 1 ], w[ i, 0 ], 0 ],
                ]
        )

    # for

    W = W_ta.stack()

    return W

# skew



"""

Library of needle shape sensing cost functions.

Author: Dimitri Lezcano

"""

from typing import Union

import numpy as np

from needle_shape_sensing import numerical, intrinsics


def constant_curvature_cost( eta: np.ndarray, data, s_m: Union[ list, np.ndarray ], ds: float,
                             L: float = None, N: int = None, weights: np.ndarray = None, scalef: float = 1 ) -> float:
    """ Perform constant curvature cost functions"""
    # determine the needle arclength components
    if N is not None:
        L = N * ds
    # N

    elif L is not None:
        pass  # no need to do anything

    # elif
    else:
        # raise AttributeError( "Either 'L' or 'N' needs to be set." )
        L = np.inf  # assume all sensing locations are valid

    # else

    curvatures = np.repmat( eta[ :2 ].reshape( 1, -1 ), data.shape[ 0 ], axis=0 )

    cost = scalef * curvature_cost( data, curvatures, np.arange( data.shape[ 0 ] ), weights=weights )

    return cost


# constant_curvature_cost

def singlebend_singlelayer_cost( eta: np.ndarray, data: np.ndarray, s_m: Union[ list, np.ndarray ], ds: float,
                                 B: np.ndarray, L: float = None, N: int = None, Binv: np.ndarray = None,
                                 R_init: np.ndarray = np.eye( 3 ), weights: np.ndarray = None, scalef: float = 1,
                                 arg_check: bool = False, continuous: bool = False ) -> float:
    """ Single bend and single layer needle cost function

        Args:
            :param eta: 4-d numpy vector of the format [kc; w_init]
            :param data: N x 3 numpy array of the curvatures at the measurement locations
            :param s_m: N-list of indexed measurement locations from tip of needle (must be integers >= 0)
            :param ds: float of the integration arclength increments
            :param B: the needle stiffness matrix
            :param L: (Default = None) float of the length of needle
            :param N: (Default = None) int of the number of points we are integrating
            :param Binv: (Default = None) inv(B)
            :param R_init: (Default = 3x3 identity) initial orientation of the needle
            :param weights: (Default = 1) the reliability weights for each of the active areas
            :param scalef: (Default = 1) the scaling for the cost function
            :param arg_check: (Default = False) whether to check if the arguments are valid
            :param continuous: (Default = False) whether to perform continuous integration or not

        Return:
            :return: cost for single bend and single-layer shape sensing

    """

    # argument checking
    if arg_check:
        pass

    # unpack the arguments and set-up
    kc = eta[ 0 ]
    w_init = eta[ 1:4 ]

    # determine the needle arclength components
    if N is not None:
        s = ds * np.arange( N ).reshape( -1, 1 )
        L = max( s )
    # N

    elif L is not None:
        N = int( L / ds ) + 1
        s = ds * np.arange( N ).reshape( -1, 1 )

    # elif
    else:
        raise AttributeError( "Either 'L' or 'N' needs to be set." )

    # else

    # determine w0 and w0prime
    k0, k0prime = intrinsics.SingleBend.k0_1layer( s, kc, L, return_callable=continuous )

    # perform integration to get wv
    if continuous:
        w0 = lambda s: np.append( k0( s ), [ 0, 0 ] )
        w0prime = lambda s: np.append( k0prime( s ), [ 0, 0 ] )

        _, _, wv = numerical.integrateEP_w0_ode( w_init, w0, w0prime, B, s, s0=0, ds=ds, R_init=R_init, Binv=Binv,
                                                 arg_check=arg_check, wv_only=True )

    # if
    else:
        w0 = np.hstack( (k0.reshape( -1, 1 ), np.zeros( (N, 2) )) )
        w0prime = np.hstack( (k0prime.reshape( -1, 1 ), np.zeros( (N, 2) )) )

        _, _, wv = numerical.integrateEP_w0( w_init, w0, w0prime, B, s0=0, ds=ds, R_init=R_init, Binv=Binv,
                                             arg_check=arg_check, wv_only=True )

    # else

    # determine AA locations
    # s_m_idx = np.argwhere( s_m.reshape( -1, 1 ) == s.ravel() )[ :, 1 ]
    s_m_idx = np.argmin(np.abs(s_m.reshape(-1,1) - s.ravel()), axis=1)

    # compute cost
    cost = scalef * curvature_cost( data, wv, s_m_idx, weights=weights )

    return cost


# singlebend_singlelayer_cost


def singlebend_doublelayer_cost( eta: np.ndarray, data: np.ndarray, s_m: Union[ list, np.ndarray ], ds: float,
                                 B: np.ndarray, L: float = None, s_crit: float = None, z_crit: float = None,
                                 N: int = None, Binv: np.ndarray = None, R_init: np.ndarray = np.eye( 3 ),
                                 weights: np.ndarray = None, scalef: float = 1, arg_check: bool = False,
                                 continuous: bool = False ) -> float:
    """ Single bend and single layer needle cost function

        Args:
            :param eta: 5-d numpy vector of the format [kc1; kc2; w_init]
            :param data: N x 3 numpy array of the curvatures at the measurement locations
            :param s_m: N-list of indexed measurement locations (must be integers >= 0)

            :param ds: float of the integration arclength increments
            :param B: the needle stiffness matrix
            :param L: (Default = None) float of the length of needle
            :param s_crit: (Default = None) the length of the needle at boundary of insertion
            :param z_crit: (Default = None) the z-critical layer boundary (will calculate s_crit)
            :param N: (Default = None) int of the number of points we are integrating
            :param Binv: (Default = None) inv(B)
            :param R_init: (Default = 3x3 identity) initial orientation of the needle
            :param weights: (Default = 1) the reliability weights for each of the active areas
            :param scalef: (Default = 1) the scaling for the cost function
            :param arg_check: (Default = False) whether to check if the arguments are valid
            :param continuous: (Default = False) whether to perform continuous integration or not

        Return:
            :return: cost for single bend and single-layer shape sensing

    """
    # argument checking
    if arg_check:
        pass

    # unpack the arguments and set-up
    kc1 = eta[ 0 ]
    kc2 = eta[ 1 ]
    w_init = eta[ 2:5 ]

    # determine the needle arclength components
    if N is not None:
        s = ds * np.arange( N ).reshape( -1, 1 )
        L = max( s )
    # N

    elif L is not None:
        N = int( L / ds ) + 1
        s = ds * np.arange( N ).reshape( -1, 1 )

    # elif
    else:
        raise AttributeError( "Either 'L' or 'N' needs to be set." )

    # else

    # determine s_crit: the length of the needle in the first layer
    if s_crit is not None:
        pass

    # if
    elif z_crit is not None:
        s_crit = intrinsics.SingleBend.determine_2layer_boundary( kc1, L, z_crit, B, w_init=w_init, s0=0, ds=ds,
                                                                  R_init=R_init, Binv=Binv, continuous=continuous )
    # elif
    else:
        raise ValueError( "Either s_crit or z_crit must be defined." )

    # else

    # determine k0 and k0prime
    if s_crit < 0:  # single-layer
        k0, k0prime = intrinsics.SingleBend.k0_1layer( s, kc1, L, return_callable=continuous )

    else:  # double-layer
        k0, k0prime = intrinsics.SingleBend.k0_2layer( s, kc1, kc2, L, s_crit, return_callable=continuous )

    # perform integration to get wv
    if continuous:
        w0 = lambda s: np.append( k0( s ), [ 0, 0 ] )
        w0prime = lambda s: np.append( k0prime( s ), [ 0, 0 ] )

        _, _, wv = numerical.integrateEP_w0_ode( w_init, w0, w0prime, B, s, s0=0, ds=ds, R_init=R_init, Binv=Binv,
                                                 arg_check=arg_check, wv_only=True )

    # if
    else:
        w0 = np.hstack( (k0.reshape( -1, 1 ), np.zeros( (N, 2) )) )
        w0prime = np.hstack( (k0prime.reshape( -1, 1 ), np.zeros( (N, 2) )) )

        _, _, wv = numerical.integrateEP_w0( w_init, w0, w0prime, B, s0=0, ds=ds, R_init=R_init, Binv=Binv,
                                             arg_check=arg_check, wv_only=True )

    # else

    # calculate cost
    s_m_idx = np.argwhere( s_m.reshape( -1, 1 ) == s.ravel() )[ :, 1 ]
    cost = scalef * curvature_cost( data, wv, s_m_idx, weights=weights )

    return cost


# singlebend_doublelayer_cost


def doublebend_singlelayer_cost( eta: np.ndarray, data: np.ndarray, s_m: Union[ list, np.ndarray ], ds: float,
                                 s_crit: float, B: np.ndarray, L: float = None, N: int = None, Binv: np.ndarray = None,
                                 R_init: np.ndarray = np.eye( 3 ), weights: np.ndarray = None, scalef: float = 1,
                                 arg_check: bool = False, continuous: bool = False ) -> float:
    """ Single bend and single layer needle cost function

        Args:
            :param eta: 4-d numpy vector of the format [kc; w_init]
            :param data: N x 3 numpy array of the curvatures at the measurement locations
            :param s_m: N-list of indexed measurement locations (must be integers >= 0)
            :param ds: float of the integration arclength increments
            :param B: the needle stiffness matrix
            :param L: (Default = None) float of the length of needle
            :param s_crit: (Default = None) the length of the needle at boundary of insertion
            :param N: (Default = None) int of the number of points we are integrating
            :param Binv: (Default = None) inv(B)
            :param R_init: (Default = 3x3 identity) initial orientation of the needle
            :param weights: (Default = 1) the reliability weights for each of the active areas
            :param scalef: (Default = 1) the scaling for the cost function
            :param arg_check: (Default = False) whether to check if the arguments are valid
            :param continuous: (Default = False) whether to perform continuous integration or not

        Return:
            :return: cost for single bend and single-layer shape sensing

    """
    # argument checking
    if arg_check:
        pass
    # if

    # unpack the arguments and set-up
    kc = eta[ 0 ]
    w_init = eta[ 1:4 ]

    # determine the needle arclength components
    if N is not None:
        s = ds * np.arange( N ).reshape( -1, 1 )
        L = max( s )
    # N

    elif L is not None:
        N = int( L / ds ) + 1
        s = ds * np.arange( N ).reshape( -1, 1 )

    # elif
    else:
        raise AttributeError( "Either 'L' or 'N' needs to be set." )

    # else

    # determine k0 and k0prime
    if s_crit < 0:  # single-layer
        k0, k0prime = intrinsics.SingleBend.k0_1layer( s, kc, L, return_callable=continuous )

    else:  # double-bend
        k0, k0prime = intrinsics.DoubleBend.k0_1layer( s, kc, L, s_crit, return_callable=continuous )

    # perform integration to get wv
    if continuous:
        w0 = lambda s: np.append( k0( s ), [ 0, 0 ] )
        w0prime = lambda s: np.append( k0prime( s ), [ 0, 0 ] )

        _, _, wv = numerical.integrateEP_w0_ode( w_init, w0, w0prime, B, s, s0=0, ds=ds, R_init=R_init, Binv=Binv,
                                                 arg_check=arg_check, wv_only=True )

    # if
    else:
        w0 = np.hstack( (k0.reshape( -1, 1 ), np.zeros( (N, 2) )) )
        w0prime = np.hstack( (k0prime.reshape( -1, 1 ), np.zeros( (N, 2) )) )

        _, _, wv = numerical.integrateEP_w0( w_init, w0, w0prime, B, s0=0, ds=ds, R_init=R_init, Binv=Binv,
                                             arg_check=arg_check, wv_only=True )

    # else

    # determine AA locations
    s_m_idx = np.argwhere( s_m.reshape( -1, 1 ) == s.ravel() )[ :, 1 ]

    # compute cost
    cost = scalef * curvature_cost( data, wv, s_m_idx, weights=weights )

    return cost


# doublebend_singlelayer_cost


def curvature_cost( data: np.ndarray, wv: np.ndarray, s_m_idx: np.ndarray, weights: np.ndarray = None ):
    """ Calculate the curvature cost"""
    if weights is None:
        weights = np.ones( data.shape[ 0 ] )
    # if

    else:
        weights = weights[ -data.shape[ 0 ]: ]

    # else

    weights = weights / np.sum( weights )

    # calculate the error
    delta = wv[ s_m_idx, 0:2 ] - data[ :, 0:2 ]
    delta_w = delta * np.reshape( weights, (-1, 1) )
    cost = np.trace( delta_w @ delta_w.T )

    return float( cost )

# curvature_cost



import torch

from needle_shape_sensing.pytorch import geometry


def integrateEP_w0(
        w_init: torch.Tensor, w0: torch.Tensor, w0prime: torch.Tensor, B: torch.Tensor,
        seq_mask: torch.Tensor, ds: float, R_init: torch.Tensor = torch.eye( 3 ),
        Binv: torch.Tensor = None, wv_only: bool = False
):
    """
    integrate Euler-Poincare equation for needle shape sensing for given intrinsic angular deformation

    Args:
            w_init: (N, 3) initial deformation vector
            w0: (N, M, 3) intrinsic angular deformation
            w0prime: (N, M, 3) d/ds w0
            B: 3 x 3 needle stiffness matrix
            seq_mask: (N, M) mask of which are valid points to use
            ds: the arclength increments
            Binv: (Default = None) inv(B) Can be provided for numerical efficiency
            R_init: (Default = 3x3 identity) SO3 matrix for initial rotation angle
            wv_only: (Default = False) whether to only integrate wv or not.

        Return:
            (N, M, 3) needle shape, (N, M, 3, 3) SO3 matrices of orientations), (N, M, 3) angular deformation), (N, M) sequence mask
            (None, None, (N, M, 3) angular deformation) if 'wv_only' is True

    """

    if Binv is None:
        Binv = torch.linalg.inv( B )

    # if

    # unpack the shapes
    N, M = w0.shape[ 0:2 ]

    # tensor-ify B and Binv
    B_T =  torch.tile(
                torch.reshape( B, (1, B.shape[ 0 ], B.shape[ 1 ]) ),
                (N, 1, 1)
    ).type(w0.dtype)
    Binv_T = torch.tile(
                torch.reshape( Binv, (1, B.shape[ 0 ], B.shape[ 1 ]) ),
                (N, 1, 1)
    ).type(w0.dtype)

    # prepare tensors
    wv_list = [w_init]

    for idx in range( 1, M ):
        seq_mask_idx = torch.unsqueeze( seq_mask[ :, idx:idx + 1 ], dim=-1 )

        # unpack veriables
        w0_im1      = torch.unsqueeze( w0[ :, idx - 1 ], dim=-1 ) # (N, 3, 1)
        w0prime_im1 = torch.unsqueeze( w0prime[ :, idx - 1 ], dim=-1 ) # (N, 3, 1)
        wv_im1      = torch.unsqueeze( wv_list[ idx - 1 ], dim=-1 ) # (N, 3, 1)

        if idx == 1:
            wv_i = (
                        wv_im1 + 1 * ds * (
                        w0prime_im1 - Binv_T @ torch.linalg.cross(
                            torch.squeeze( wv_im1, dim=-1 ),
                            torch.squeeze( B_T @ (wv_im1 - w0_im1), dim=-1 )
                        )[ :, :, None ]
                )
                ) * seq_mask_idx.type( w0.dtype )
            
        # if
        else:
            wv_im2 = torch.unsqueeze( wv_list[ idx - 2 ], dim=-1 ) # (N, 3, 1)
            wv_i   = (
                        wv_im2 + 2 * ds * (
                        w0prime_im1 - Binv_T @ torch.linalg.cross(
                            torch.squeeze( wv_im1, dim=-1 ),
                            torch.squeeze( B_T @ (wv_im1 - w0_im1), dim=-1 )
                        )[ :, :, None ]
                )
                ) * seq_mask_idx.type( w0.dtype )
            
        # else

        wv_list.append(torch.squeeze(wv_i, dim=-1))

    # for

    wv = torch.cat([wv_i[:, None] for wv_i in wv_list], dim=1)

    if wv_only:
        pmat, Rmat = None, None

    else:
        pmat, Rmat, seq_mask = integratePose_wv( wv, seq_mask, ds=ds, R_init=R_init, )

    return pmat, Rmat, wv, seq_mask


# integrateEP_w0

def integratePose_wv(
        wv: torch.Tensor, seq_mask: torch.Tensor, ds: float, R_init=torch.eye( 3 )
):
    """ Integrate angular deformation to get the pose of the needle along it's arclengths

        :param wv: (N, M, 3) angular deformation vector
        :param seq_mask: (N, M) sequence mask for which are viable
        :param ds: (Default = None) the arclength increments desired
        :param R_init: (Default = numpy.eye(3)) Rotation matrix of the inital pose

        :returns: pmat, Rmat, sequence mask
            - pmat: (N, M, 3) position for the needle shape points in-tissue
            - Rmat: (N, M, 3, 3) SO(3) rotation matrices for
            - sequence mask: (N, M) of boolean for which sequences are viable
    """
    N, M = wv.shape[ 0:2 ]

    # initial conditions
    pmat_list = [torch.zeros((N, 3), dtype=wv.dtype, device=wv.device)]
    Rmat_list = [R_init[None].tile((N, 1, 1)).type(wv.dtype).to(wv.device)]

    for idx in range( 1, M ):
        # unpack vars
        seq_mask_i = seq_mask[ :, idx ]  # (N, )
        Rmat_im1 = Rmat_list[idx-1]   # (N, 3, 3)
        pmat_im1 = pmat_list[idx - 1] # (N, 3)

        # integrate
        Rmat_i = Rmat_im1 @ (
                geometry.exp2r(
                        ds * torch.mean( wv[ :, idx - 1:idx ], dim=1 )
                        ) * seq_mask_i[ :, None, None ].type( wv.dtype )
        ).type(Rmat_im1.dtype)

        # add results for Rmat
        Rmat_list.append(Rmat_i)

        if idx == 1:
            pmat_i = pmat_im1 + Rmat_i[ :, :, 2 ] * ds

        else:
            e3vec = torch.cat(
                    [ R[:, None] for R in Rmat_list[0:idx + 1] ],
                    dim=1
            )[:, :, :, 2]  # grab z-directions
            pmat_i = simpson_vec_int( e3vec, ds, seq_mask[ :, :idx + 1 ] )

        # add results for pmat
        pmat_list.append(pmat_i)

    # for

    pmat = torch.cat(
            [p[:, None] for p in pmat_list],
            dim=1,
    )
    Rmat = torch.cat(
            [R[:, None] for R in Rmat_list],
            dim=1
    )

    return pmat, Rmat, seq_mask


# integratePose_wv


def simpson_vec_int( f: torch.Tensor, dx: float, seq_mask: torch.Tensor ):
    """ Implementation of Simpson vector integration for tensor integration

            Original Author (MATLAB): Jin Seob Kim
            Translated Author: Dimitri Lezcano

            Args:
                 f:  (N, M, D) tensor where D is the dimension of the vector is the dimension of the parameter ( N > 2 )
                        and M Integration samples
                 dx: float of the step size
                 seq_mask: (N, M) boolean mask for which parts of the integration is valid

            Return:
                integrated tensor of shape (N, D)

    """
    N, M, D = f.shape[ :3 ]
    num_intervals = M - 1
    assert (num_intervals > 1)

    f_masked = f * seq_mask[ :, :, None ].type( f.dtype )

    # perform the integration
    int_res = torch.zeros( (f.shape[ 0 ], f.shape[ 2 ]), dtype=f.dtype, device=f.device )
    if num_intervals == 2:
        int_res = dx / 3 * torch.sum(
                f_masked[ :, 0:3 ]
                * torch.reshape( torch.tensor([ 1., 4., 1. ], device=f.device), (1, -1, 1) ).type( f_masked.dtype ),
                dim=1
        )

        return int_res

    # if

    elif num_intervals == 3:
        int_res = 3 / 8 * dx * torch.sum(
                f_masked[ :, 0:4 ]
                * torch.reshape( torch.tensor([ 1., 3., 3., 1. ], device=f.device), (1, -1, 1) ).type( f_masked.dtype ),
                dim=1
        )

        return int_res

    # elif

    if num_intervals % 2 != 0:
        int_res = (
                int_res +
                3 / 8 * dx * torch.sum(
                    f_masked[ :, -4: ]
                    * torch.reshape( torch.tensor([ 1., 3., 3., 1. ], device=f.device), (1, -1, 1) ).type( f_masked.dtype ),
                    dim=1
                )
        )
        m = num_intervals - 3

    # if
    else:
        m = num_intervals

    # else

    int_res = (
            int_res
            + dx / 3 * (
                    f_masked[ :, 0 ] + f_masked[ :, m ]
                    + 4 * torch.sum( f_masked[ :, 1:m:2 ], dim=1 )
            )
    )

    if m > 2:
        int_res = (
                int_res
                + 2 / 3 * dx * torch.sum( f_masked[ :, 2:m:2 ], dim=1 )
        )
    # if

    return int_res

# simpson_vec_int



"""

Library of numerical computations for needle shape sensing

Author: Dimitri Lezcano

"""
import warnings
from typing import Union, Callable

import numpy as np
import scipy.integrate.odepack
import scipy.optimize as spoptim
from scipy.integrate import odeint, solve_ivp
from scipy import interpolate
from spatialmath.base import exp2r
from numba import jit

from needle_shape_sensing import geometry, cost_functions
from needle_shape_sensing.sensorized_needles import FBGNeedle


class NeedleParamOptimizations:
    def __init__(
            self, fbgneedle: FBGNeedle, ds: float = 0.5, optim_options: dict = None,
            continuous: bool = True ):
        assert (ds > 0)
        self.fbg_needle = fbgneedle

        # don't calculate each time
        self.__needle_B = self.fbg_needle.B
        self.__needle_Binv = np.linalg.inv( self.fbg_needle.B )

        self.ds = ds

        # optimizer options
        default_options = {
            'w_init_bounds': [ (-0.01, 0.01) ] * 3,
            'kc_bounds'    : [ (0, 0.01) ],
            'tol'          : 1e-8,
            'method'       : 'SLSQP',
            }
        self.options = default_options
        self.options.update( optim_options if optim_options is not None else { } )

        # integration options
        self.continuous = continuous

    # __init__

    def constant_curvature( self, data: np.ndarray, L: float ):
        """
            Compute the constant curvature

        """
        assert (L > 0)
        # check for inserted into tissue
        s_data_ins, inserted_sensors = self.fbg_needle.calculate_length_measured(
                L, tip=True, valid=True )
        data_ins = data[ inserted_sensors ]
        if len( self.fbg_needle.weights ) > 0:
            weights = np.array(
                    [ self.fbg_needle.weights[ key ] for (key, inserted) in
                      zip( self.fbg_needle.weights.keys(), inserted_sensors ) if inserted ] )
        # if
        else:
            weights = np.ones_like( s_data_ins )

        # else

        # normalize the weights
        weights = weights / np.sum( weights )

        # compute the weighted-mean curvature from the measurements
        data_ins_w = data_ins * weights.reshape( -1, 1 )
        curvature = data_ins_w.sum( axis=0 )

        return curvature

    # constant_curvature

    def singlebend_singlelayer_k0(
            self, k_c_0: float, w_init_0: np.ndarray, data: np.ndarray, L: float,
            R_init: np.ndarray = np.eye( 3 ), **kwargs ):
        """
            :param k_c_0: initial intrinsic curvature
            :param w_init_0: initial w_init
            :param data: N x 3 array of curvatures for each of the active areas
            :param L: the insertion depth
            :param R_init: (Default = numpy.eye(3)) the initial rotation offset

            :keyword: any scipy optimize (least_squares) kwargs

            :return: kc, w_init optimized, optimization results
        """
        # argument checking
        assert (L > 0)
        # check for inserted into tissue
        s_data_ins, inserted_sensors = self.fbg_needle.calculate_length_measured(
                L, tip=True, valid=True )
        data_ins = data[ inserted_sensors ]
        if len( self.fbg_needle.weights ) > 0:
            weights = np.array(
                    [ self.fbg_needle.weights[ key ] for (key, inserted) in
                      zip( self.fbg_needle.weights.keys(), inserted_sensors ) if inserted ] )
        # if
        else:
            weights = None

        # else

        # perform the optimization
        eta_0 = np.append( [ k_c_0 ], w_init_0 )
        Binv = kwargs.get( 'Binv', self.__needle_Binv )
        cost_kwargs = {
                'L'      : L,
                'Binv'   : Binv,
                'R_init' : R_init,
                'weights': weights
                }
        c_0 = cost_functions.singlebend_singlelayer_cost(
                eta_0, data_ins, s_data_ins, self.ds, self.__needle_B,
                scalef=1, arg_check=True, **cost_kwargs )
        c_f = 1 / c_0 if c_0 > 0 else 1
        cost_fn = lambda eta: cost_functions.singlebend_singlelayer_cost(
            eta=eta,
            data=data_ins,
            s_m=s_data_ins,
            ds=self.ds,
            B=self.fbg_needle.B,
            scalef=c_f,
            arg_check=False,
            continuous=self.continuous,
            **cost_kwargs
        )
        res = self.__optimize( cost_fn, eta_0, **kwargs )
        kc, w_init = res.x[ 0 ], res.x[ 1:4 ]

        return kc, w_init, res

    # singlebend_singlelayer_k0

    def singlebend_doublelayer_k0(
            self, kc1_0: float, kc2_0: float, w_init_0: np.ndarray, data: np.ndarray, L: float,
            s_crit: float = None, z_crit: float = None, R_init: np.ndarray = np.eye( 3 ),
            **kwargs ):
        """
            :param kc1_0: initial intrinsic curvature for layer 1
            :param kc2_0: initial intrinsic curvature for layer 2
            :param w_init_0: initial w_init
            :param data: N x 3 array of curvatures for each of the active areas
            :param L: the insertion depth
            :param z_crit: (Default = None) the length of the first layer
            :param s_crit: (Default = None) the length of the needle in first layer
            :param R_init: (Default = numpy.eye(3)) the initial rotation offset

            :keyword: any scipy optimize (least_squares) kwargs

            :return: kc1, kc2, w_init optimized, optimization results
        """
        assert (L >= z_crit > 0)
        # check for inserted into tissue
        s_data_ins, inserted_sensors = self.fbg_needle.calculate_length_measured(
                L, tip=True, valid=True )
        data_ins = data[ inserted_sensors ]
        if len( self.fbg_needle.weights ) > 0:
            weights = np.array(
                    [ self.fbg_needle.weights[ key ] for (key, inserted) in
                      zip( self.fbg_needle.weights.keys(), inserted_sensors ) if inserted ] )
        # if
        else:
            weights = None

        # else

        # perform the optimization
        eta_0 = np.append( [ kc1_0, kc2_0 ], w_init_0 )
        Binv = kwargs.get( 'Binv', self.__needle_Binv )
        cost_kwargs = {
                'L'      : L,
                's_crit' : s_crit,
                'z_crit' : z_crit,
                'Binv'   : Binv,
                'R_init' : R_init,
                'weights': weights
                }
        c_0 = cost_functions.singlebend_doublelayer_cost(
                eta_0, data_ins, s_data_ins, self.ds, self.__needle_B,
                scalef=1, arg_check=True, **cost_kwargs )
        c_f = 1 / c_0 if c_0 > 0 else 1
        cost_fn = lambda eta: cost_functions.singlebend_doublelayer_cost(
                eta, data_ins, s_data_ins, self.ds,
                self.fbg_needle.B,
                scalef=c_f, arg_check=False,
                continuous=self.continuous,
                **cost_kwargs )
        res = self.__optimize( cost_fn, eta_0, **kwargs )
        kc1, kc2, w_init = res.x[ 0 ], res.x[ 1 ], res.x[ 2:5 ]

        return kc1, kc2, w_init, res

    # singlebend_doublelayer_k0

    def doublebend_singlelayer_k0(
            self, kc_0: float, w_init_0: np.ndarray, data: np.ndarray, L: float,
            s_crit: float, R_init: np.ndarray = np.eye( 3 ), **kwargs ):
        """
            :param kc_0: initial intrinsic curvature for layer 1
            :param w_init_0: initial w_init
            :param data: N x 3 array of curvatures for each of the active areas
            :param L: the insertion depth
            :param s_crit: (Default = None) the length of the needle in first layer
            :param R_init: (Default = numpy.eye(3)) the initial rotation offset

            :keyword: any scipy optimize (least_squares) kwargs

            :return: kc1, kc2, w_init optimized, optimization results
        """
        assert (L >= s_crit > 0)
        # check for inserted into tissue
        s_data_ins, inserted_sensors = self.fbg_needle.calculate_length_measured(
                L, tip=True, valid=True )
        data_ins = data[ inserted_sensors ]
        if len( self.fbg_needle.weights ) > 0:
            weights = np.array(
                    [ self.fbg_needle.weights[ key ] for (key, inserted) in
                      zip( self.fbg_needle.weights.keys(), inserted_sensors ) if inserted ] )
        # if
        else:
            weights = None

        # else

        # perform the optimization
        eta_0 = np.append( [ kc_0 ], w_init_0 )
        Binv = kwargs.get( 'Binv', self.__needle_Binv )
        cost_kwargs = {
                'L'      : L,
                'Binv'   : Binv,
                'R_init' : R_init,
                'weights': weights
                }
        c_0 = cost_functions.doublebend_singlelayer_cost(
                eta_0, data_ins, s_data_ins, self.ds, s_crit,
                self.fbg_needle.B,
                scalef=1, arg_check=True, **cost_kwargs )
        c_f = 1 / c_0 if c_0 > 0 else 1
        cost_fn = lambda eta: cost_functions.doublebend_singlelayer_cost(
                eta, data_ins, s_data_ins, self.ds,
                s_crit, self.__needle_B,
                scalef=c_f, arg_check=False,
                continuous=self.continuous,
                **cost_kwargs )
        res = self.__optimize( cost_fn, eta_0, **kwargs )

        kc, w_init = res.x[ 0 ], res.x[ 1:4 ]

        return kc, w_init, res

    # doublebend_singlelayer_k0

    def __optimize( self, cost_fn, eta_0, **kwargs ):
        """ Optimize wrapper for multiple functions"""
        optim_options = self.options.copy()
        optim_options.update( kwargs )

        # filter specific bounds and add bounds if not already specified
        exclude_keys = [ 'w_init_bounds', 'kc_bounds', 'needle_rotations' ]
        optim_options = dict(
            filter(
                lambda x: x[ 0 ] not in exclude_keys,
                optim_options.items()
            )
        )
        if optim_options.get( 'bounds' ) is None:
            bounds = (
                self.options[ 'kc_bounds' ] * (eta_0.size - 3) 
                + self.options[ 'w_init_bounds' ]
            )
            optim_options[ 'bounds' ] = bounds

        # bounds

        # perform optimization
        with warnings.catch_warnings():
            warnings.simplefilter( "ignore", scipy.integrate.odepack.ODEintWarning )
            warnings.simplefilter( "ignore", RuntimeWarning )
            result = spoptim.minimize( cost_fn, eta_0, **optim_options )

        # with

        return result

    # __optimize


# NeedleParamOptimizations

def compute_orientation_from_shape(shape: np.ndarray):
    """ Uses the shape to compute the orientation using the tangent, normal, and binormal vectors
        from differentiable curves

    Args:
        shape: N x 3 shape matrix for computing the orientation matrix

    Returns:
        orientation: N x 3 x 3 orientation matrix

    """
    d_shape = np.diff(shape, 1, 0)
    ds = np.linalg.norm(
        d_shape,
        ord=2,
        axis=1,
        keepdims=True,
    )

    tangent_vect = d_shape / ds
    tangent_vect = np.concatenate(
        (
            np.array([[0, 0, 1]]),
            tangent_vect
        ),
        axis=0
    )
    tangent_vect = normalize_orientations(tangent_vect)

    normal_vect = np.diff(tangent_vect, 1, 0) / ds
    normal_vect = np.concatenate(
        (
            np.array([[1, 0, 0]]),
            normal_vect
        ),
        axis=0
    )
    normal_vect = normalize_orientations(normal_vect)

    binormal_vect = np.cross(tangent_vect, normal_vect)
    binormal_vect = normalize_orientations(binormal_vect)

    orientation = np.stack(
        (
            normal_vect,
            binormal_vect,
            tangent_vect,
        ),
        axis=2
    )

    return orientation

# comptue_orientation_from_shape


@jit( nopython=True, parallel=True )
def jit_linear_interp1d( t: float, x_all: np.ndarray, t_all: np.ndarray, is_sorted: bool = False ):
    """ numba 1d linear interpolation"""
    if not is_sorted:
        idxs = np.argsort( t_all )
        t_all = t_all[ idxs ]
        x_all = x_all[ idxs ]

    # if
    if t in t_all:
        return x_all[ t == t_all ]

    else:
        t1_idx = np.argmin( np.abs( t - t_all ) )
        t1 = t_all[ t1_idx ]
        x1 = x_all[ t1 ]

        t2_idx = t1_idx + 1 if t1 < t else t1_idx - 1
        t2 = t_all[ t2_idx ]
        x2 = x_all[ t2 ]

        return x1 + (x2 - x1) * (t - t1) / (t2 - t1)

    # else


# jit_linear_interp1d

# @jit( nopython=True, parallel=True )
def differential_EPeq( wv: np.ndarray, s: float, w0, w0prime, B: np.ndarray, Binv: np.ndarray ):
    """ The differential form of the Euler-Poincare equation equation

        :param wv: the current omega v
        :param s: the current arclength
        :param w0: the intrinsic angular deformation
        :param w0prime: d(w0)/ds
        :param B: the stiffness matrix
        :param Binv: the inverse of the stiffness matrix

        :returns: dwv/ds

    """
    dwvds = w0prime( s ).ravel() - Binv @ np.cross( wv, B @ (wv - w0( s ).ravel()) )

    return dwvds


# differential_EPeq

def integrateEP_w0(
        w_init: np.ndarray, w0: np.ndarray, w0prime: np.ndarray, B: np.ndarray,
        s: np.ndarray = None, s0: float = 0, ds: float = None, R_init: np.ndarray = np.eye( 3 ),
        Binv: np.ndarray = None, needle_rotations: list = None, arg_check: bool = True,
        wv_only: bool = False ) -> (
        np.ndarray, np.ndarray, np.ndarray):
    """ integrate Euler-Poincare equation for needle shape sensing for given intrinsic angular deformation

        Original Author: Jin Seob Kim
        Editing Author: Dimitri Lezcano

        Args:
            w_init: 3-D initial deformation vector
            w0: N x 3 intrinsic angular deformation
            w0prime: N x 3 d/ds w0
            B: 3 x 3 needle stiffness matrix
            s: (Default = None) the arclengths desired (Not implemented)
            s0: (Default = 0) the initial length to start off with
            ds: (Default = None) the arclength increments desired
            Binv: (Default = None) inv(B) Can be provided for numerical efficiency
            R_init: (Default = 3x3 identity) SO3 matrix for initial rotation angle
            needle_rotations: (Default = None) list of rotations in degrees of the needle rotations
            arg_check: (Default = False) whether to check if the arguments are valid
            wv_only: (Default = False) whether to only integrate wv or not.

        Return:
            (N x 3 needle shape, N x 3 x 3 SO3 matrices of orientations), N x 3 angular deformation)
            (None, None, N x 3 angular deformation) if 'wv_only' is True
    """
    # argument validation
    if arg_check:
        assert (w_init.size == 3)
        w_init = w_init.flatten()

        assert (w0.shape[ 1 ] == 3 and w0prime.shape[ 1 ] == 3)
        assert (w0.shape[ 0 ] == w0prime.shape[ 0 ])

        assert (B.shape == (3, 3))
        assert (geometry.is_SO3( R_init ))

        assert (s0 >= 0)

    # if

    # argument parsing
    N = w0.shape[ 0 ]
    if (s is None) and (ds is not None):
        s = s0 + ds * np.arange( N )

    elif s is not None:
        if np.any( np.diff( s )[ 0 ] != np.diff( s ) ):  # ensure constant ds
            raise NotImplementedError( 'You must use ds here only for uniform distribution' )

        else:
            ds = np.diff( s )[ 0 ]

    # elif

    else:
        raise ValueError( "Either `s` or `ds` needs to be provided. One or the other." )

    s = s[ s >= s0 ]

    if Binv is None:
        Binv = np.linalg.inv( B )

    elif arg_check:
        assert (Binv.shape == (3, 3))

    # needle rotation axis
    if needle_rotations is None:
        needle_rotations = [ 0 ] * len( s )

    else:
        needle_rotations = needle_rotations[ s >= s0 ]

    # update w0 and w0prime
    for idx in range( N ):
        Rz = geometry.rotz( needle_rotations[ idx ] )
        w0[ idx ] = Rz @ w0[ idx ]
        w0prime[ idx ] = Rz @ w0prime[ idx ]

    # for

    # initialize the return matrices
    wv = np.zeros( (N, 3) )

    # initial conditions
    wv[ 0 ] = w_init

    # perform integration to calculate angular deformation vector
    for idx in range( 1, N ):
        ds = s[ idx ] - s[ idx - 1 ]  # calculate ds
        Rz = geometry.rotz( needle_rotations[ idx ] )
        if idx == 1:
            wv[ idx ] = w_init + ds * (
                    w0prime[ 0 ] - Binv @ np.cross( w_init, B @ (w_init - w0[ 0 ]) ))

        else:
            wv[ idx ] = wv[ idx - 2 ] + 2 * ds * (
                    w0prime[ idx - 1 ] - Binv @ np.cross(
                    wv[ idx - 1 ], B @ (wv[ idx - 1 ] - w0[ idx - 1 ]) ))

    # for

    # integrate angular deviation vector in order to get the pose
    if wv_only:
        pmat, Rmat = None, None
    else:
        pmat, Rmat = integratePose_wv( wv, s=s, s0=s0, ds=ds, R_init=R_init )

    return pmat, Rmat, wv


# integrateEP_w0

def integrateEP_w0_ode(
        w_init: np.ndarray, w0: Union[ Callable, np.ndarray ],
        w0prime: Union[ Callable, np.ndarray ],
        B: np.ndarray, s: np.ndarray, s0: float = 0, ds: float = None,
        R_init: np.ndarray = np.eye( 3 ), Binv: np.ndarray = None, arg_check: bool = True,
        needle_rotations: list = None, wv_only: bool = False, integration_method = 'RK45' ) -> (
        np.ndarray, np.ndarray, np.ndarray):
    """ integrate Euler-Poincare equation for needle shape sensing for given intrinsic angular deformation
        using scipy.integrate

        Author: Dimitri Lezcano

        Args:
            w_init: 3-D initial deformation vector
            w0: Callable function or N x 3 intrinsic angular deformation
            w0prime: Callable function or N x 3 d/ds w0
            B: 3 x 3 needle stiffness matrix
            s: the arclengths desired (Not implemented)
            s0: (Default = 0) the initial length to start off with
            ds: (Default = None) the arclength increments desired
            Binv: (Default = None) inv(B) Can be provided for numerical efficiency
            R_init: (Default = 3x3 identity) SO3 matrix for initial rotation angle
            arg_check: (Default = False) whether to check if the arguments are valid
            needle_rotations: (Default = None) a list of needle axis rotations
            wv_only: (Default = False) whether to only integrate wv or not.
            integration_method: (Default = 'RK45') type of integration ('odeint', 'RK23', 'RK45')
                'odeint': speed but unstable
                'RK23':   slower than 'odeint', faster than 'RK45', more stable than 'odeint'
                'RK45':   slowest, but more stable than the rest.

        Return:
            (N x 3 needle shape, N x 3 x 3 SO3 matrices of orientations), N x 3 angular deformation)
            (None, None, N x 3 angular deformation) if 'wv_only' is True
    """
    if arg_check:
        assert (w_init.size == 3)
        w_init = w_init.flatten()

        assert (B.shape == (3, 3))
        assert (geometry.is_SO3( R_init ))

        assert (s0 >= 0)

    # if

    # argument parsing
    s = s[ s >= s0 ]

    if Binv is None:
        Binv = np.linalg.inv( B )

    elif arg_check:
        assert (Binv.shape == (3, 3))

    # setup intrinsic curvature functions
    if callable( w0 ):
        w0_fn = w0
    else:
        w0_fn = interpolate.interp1d( s, w0.T, fill_value='extrapolate' )
        # w0_fn = lambda t: jit_linear_interp1d( t, w0, s )

    if callable( w0prime ):
        w0prime_fn = w0prime
    else:
        w0prime_fn = interpolate.interp1d( s, w0prime.T, fill_value='extrapolate' )
        # w0prime_fn = lambda t: jit_linear_interp1d( t, w0prime, s )

    # use needle rotation interpolation
    if needle_rotations is None:
        needle_rotation_fn = lambda s: 0  # default nullify implementation

    elif callable( needle_rotations ):
        needle_rotation_fn = needle_rotations

    else:
        needle_rotation_fn = interpolate.interp1d( s, needle_rotations, fill_value=0 )

    w0_rot_fn = lambda s: geometry.rotz( needle_rotation_fn( s ) ) @ w0_fn( s )
    w0prime_rot_fn = lambda s: geometry.rotz( needle_rotation_fn( s ) ) @ w0prime_fn( s )

    # perform integration
    ode_EP = lambda s, wv: differential_EPeq( wv, s, w0_rot_fn, w0prime_rot_fn, B, Binv )
    if integration_method.lower() == 'odeint':
        wv = odeint( ode_EP, w_init, s, full_output=False, hmin=ds / 2, h0=ds / 2, tfirst=True )

    else:
        wv_res = solve_ivp( ode_EP, (s0, s.max()), w_init, method=integration_method, t_eval=s,
                    first_step=ds )  # 'RK23' for speed (all slower than odeint) 'RK45' for accuracy
        wv = wv_res.y.T

    # else


    # integrate angular deviation vector in order to get the pose
    if wv_only:
        pmat, Rmat = None, None
    else:
        pmat, Rmat = integratePose_wv( wv, s=s, s0=s0, ds=ds, R_init=R_init )

    return pmat, Rmat, wv


# integrateEP_w0_ode

def integratePose_wv(
        wv, s: np.ndarray = None, s0: float = 0, ds: float = None,
        R_init: np.ndarray = np.eye( 3 ) ):
    """ Integrate angular deformation to get the pose of the needle along it's arclengths

        :param wv: N x 3 angular deformation vector
        :param s: numpy array of arclengths to integrate
        :param ds: (Default = None) the arclength increments desired
        :param s0: (Default = 0) the initial length to start off with
        :param R_init: (Default = numpy.eye(3)) Rotation matrix of the inital pose

        :returns: pmat, Rmat
            - pmat: N x 3 position for the needle shape points in-tissue
            -Rmat: N x 3 x 3 SO(3) rotation matrices for
    """
    # set-up the containers
    N = wv.shape[ 0 ]
    pmat = np.zeros( (N, 3) )
    Rmat = np.expand_dims( np.eye( 3 ), axis=0 ).repeat( N, axis=0 )
    Rmat[ 0 ] = R_init

    # get the arclengths
    if (s is None) and (ds is not None):
        s = s0 + np.arange( N ) * ds
    elif s is not None:
        pass
    else:
        raise ValueError( "Either 's' or 'ds' must be used, not both." )

    # else

    # integrate angular deviation vector in order to get the pose
    for i in range( 1, N ):
        ds = s[ i ] - s[ i - 1 ]
        Rmat[ i ] = Rmat[ i - 1 ] @ exp2r( ds * np.mean( wv[ i - 1:i ], axis=0 ) )

        e3vec = Rmat[ :i + 1, :, 2 ].T  # grab z-direction coordinates

        if i == 1:
            pmat[ i ] = pmat[ i - 1 ] + Rmat[ i, :, 2 ] * ds
        else:
            pmat[ i ] = simpson_vec_int( e3vec, ds )

    # for

    return pmat, Rmat


# integratePose_wv

def interpolate_curve_s( points: np.ndarray, s_interp: np.ndarray, axis=0 ):
    """ Interpolate a curve based on its arclength via linear interpolation

        :param points: the points to interpolate by (N x M array)
        :param s_interp: 1D array of arclengths to interpolate
        :param axis: (Default=0), the axis to perform the interpolation on.

        :return: Interpolated curve points w.r.t s_interp, s_interp


    """
    # get the arclengths of the curve
    L, s = geometry.arclength( points, axis=axis )

    s_interp = s_interp[
        (s_interp <= L) & (s_interp >= s.min()) ]  # cut-off extrapolated curve values

    # begin interpolation
    points_interp = interpolate.interp1d( s, points, axis=axis )( s_interp )

    return points_interp, s_interp


# interpolate_curve_s

def normalize_orientations(vects: np.ndarray):
    """ Normalizes the orientation vectors

    If vects[i] == [0, 0, 0], will find first non-zero vector vects[j] s.t. j < i

    Args:
        vects: N x 3 array of N, 3D vectors

    Returns:
        normalized_vects: N x 3 array of normalized 3D vectors

    """

    norms = np.linalg.norm(vects, ord=2, axis=1, keepdims=True)

    normalized_vects  = vects.copy()
    mask_nz_zero_vect = np.ravel(norms > 0)

    normalized_vects[mask_nz_zero_vect] /= norms[mask_nz_zero_vect]

    if np.all(mask_nz_zero_vect):
        return normalized_vects

    # if

    remappings = np.arange(normalized_vects.shape[0])
    while np.any(norms[remappings] == 0):
        remappings[np.ravel(norms[remappings] == 0)] -= 1

    # while

    normalized_vects = normalized_vects[remappings]

    return normalized_vects

# normalize_orientations


def simpson_vec_int( f: np.ndarray, dx: float ) -> np.ndarray:
    """ Implementation of Simpson vector integration

        Original Author (MATLAB): Jin Seob Kim
        Translated Author: Dimitri Lezcano

        Args:
             f:  m x n numpy array where m is the dimension of the vector and n is the dimension of the parameter ( n > 2 )
                    Integration intervals
             dx: float of the step size

        Return:
            numpy vector of shape (m,)

    """
    num_intervals = f.shape[ 1 ] - 1
    assert (num_intervals > 1)  # need as least a single interval

    # TODO: non-uniform dx integration

    # perform the integration
    if num_intervals == 2:  # base case 1
        int_res = dx / 3 * np.sum( f[ :, 0:3 ] * [ [ 1, 4, 1 ] ], axis=1 )
        return int_res

    # if
    elif num_intervals == 3:  # base case 2
        int_res = 3 / 8 * dx * np.sum( f[ :, 0:4 ] * [ [ 1, 3, 3, 1 ] ], axis=1 )
        return int_res

    # elif

    else:
        int_res = np.zeros( (f.shape[ 0 ]) )

        if num_intervals % 2 != 0:
            int_res += 3 / 8 * dx * np.sum( f[ :, -4: ] * [ [ 1, 3, 3, 1 ] ], axis=1 )
            m = num_intervals - 3

        # if
        else:
            m = num_intervals

        # else

        int_res += dx / 3 * (f[ :, 0 ] + 4 * np.sum( f[ :, 1:m:2 ], axis=1 ) + f[ :, m ])

        if m > 2:
            int_res += dx / 3 * 2 * np.sum( f[ :, 2:m:2 ], axis=1 )

        # if

    # else

    return int_res

# simpson_vec_int



"""

Library of needle shape sensing intrinic measurement functions

Author: Dimitri Lezcano

"""

from enum import Flag
from typing import Union, Callable

import numpy as np

from needle_shape_sensing import numerical, geometry


class SHAPETYPE( Flag ):
    # first byte is for the number of bends, the second byte is for the number of layers ( - 1)
    CONSTANT_CURVATURE = 0x00
    SINGLEBEND_SINGLELAYER = 0x01
    SINGLEBEND_DOUBLELAYER = 0x02
    DOUBLEBEND_SINGLELAYER = 0x10 | SINGLEBEND_SINGLELAYER

    def is_singlebend(self):
        return self.value & 0x10 == 0x00
    
    # is_singlebend

    def is_doublebend(self):
        return self.value & 0x10 == 0x10
    
    # is_doublebend

    def is_singlelayer(self):
        return self.value & 0x01 == 0x01
    
    # is_singlelayer

    def is_doublelayer(self):
        return self.value & 0x02 == 0x02
    
    # is_doublelayer

    def get_k0( self ):
        """ Get the intrinsics kappa 0 function

        """
        k0_fns = {
                SHAPETYPE.CONSTANT_CURVATURE    : ConstantCurvature.k0,
                SHAPETYPE.SINGLEBEND_SINGLELAYER: SingleBend.k0_1layer,
                SHAPETYPE.SINGLEBEND_DOUBLELAYER: SingleBend.k0_2layer,
                SHAPETYPE.DOUBLEBEND_SINGLELAYER: DoubleBend.k0_1layer,
        }

        return k0_fns[ self ]

    # get_k0

    def get_shape_class( self ):
        """ Get the intrinsics shape class

        """
        classes = {
                SHAPETYPE.CONSTANT_CURVATURE    : ConstantCurvature,
                SHAPETYPE.SINGLEBEND_SINGLELAYER: SingleBend,
                SHAPETYPE.SINGLEBEND_DOUBLELAYER: SingleBend,
                SHAPETYPE.DOUBLEBEND_SINGLELAYER: DoubleBend,
        }

        return classes[ self ]

    # get_shape_class

    def k0( self, *args, **kwargs ):
        """ Get kappa_0 """
        return self.get_k0()( *args, **kwargs )

    # k0

    def w0( self, *args, **kwargs ):
        """ Get omega_o intrinsic 3D curvature function """
        k0 = self.k0(*args, **kwargs)
        if callable( k0 ):
            w0 = lambda s: np.array( [ k0( s ), 0, 0 ], dtype=np.float64 )

        # if
        else:
            w0 = np.vstack(
                    (k0, np.zeros( (2, k0.size) )),
                    dtype=k0.dtype
            )

        # else

        return w0

    # w0


# enum class: SHAPETYPES

class AirDeflection:
    @staticmethod
    def shape( s: np.ndarray, insertion_point: np.ndarray, cubic_fit: bool = False ):
        """
            Get the shape from in-air deflection.

            :param s: Arclengths of the needle to consider (cut-off so within insertion point region)
            :param insertion_point: the point (relative to needle base coordinate frame (0,0,0) where insertion is
            :param cubic_fit: (bool, Default=False), whether to use cubic fitting or rollback to quadratic fit.

            :return: numpy array of 3D points of the needle deflected in air.

        """
        if cubic_fit:  # mechanical modelling fit
            points = AirDeflection.shape_cubic( s, insertion_point )

        else:  # quadratic fit
            points = AirDeflection.shape_quadratic( s, insertion_point )

        Rmat = numerical.compute_orientation_from_shape(points)

        return points, Rmat

    # shape

    @staticmethod
    def shape_cubic( s: np.ndarray, insertion_point: np.ndarray ):
        """
            Get the shape from in-air deflection using cubic fit from mechanical modelling.

            :param s: Arclengths of the needle to consider (cut-off so within insertion point region)
            :param insertion_point: the point (relative to needle base coordinate frame (0,0,0) where insertion is

            :return: numpy array of 3D points of the needle deflected in air.

        """
        # length of entry point approximation # TODO: update for more accurate modelling
        L_entry = np.linalg.norm( insertion_point )

        # fit beam mechanics-based cubic polynomial based on insertion point
        z_ins = insertion_point[ 2 ]
        a_xy = insertion_point[ 0:2 ] / (z_ins ** 3 - 3 * L_entry * z_ins ** 2)

        # generate z-poitns
        s = np.unique( s )
        dz = np.min( np.abs( np.diff( s ) ) )
        z = np.linspace( 0, z_ins, int( z_ins // dz ) + 1 )

        # compute points along z-axis
        points_z = np.hstack((
            a_xy.reshape( 1, -1 ) * (z ** 3 - 3 * L_entry * z_ins ** 2).reshape( -1, 1 ),
            z.reshape( -1, 1 )
        ))

        L, _ = geometry.arclength( points_z, axis=0 )

        # interpolate to get s
        s = s[ s <= L ]  # cut-off extrapolated features
        points, _ = numerical.interpolate_curve_s( points_z, s, axis=0 )

        return points

    # shape_cubic

    @staticmethod
    def shape_quadratic( s: np.ndarray, insertion_point: np.ndarray ):
        """
            Get the shape from in-air deflection using quadratic fit.

            :param s: Arclengths of the needle to consider (cut-off so within insertion point region)
            :param insertion_point: the point (relative to needle base coordinate frame (0,0,0) where insertion is

            :return: numpy array of 3D points of the needle deflected in air.

        """
        # fit parabola based on insertion point
        z_ins = insertion_point[ 2 ]
        a_xy = insertion_point[ 0:2 ] / z_ins ** 2
        a_xy_norm = np.linalg.norm( a_xy )

        # determine z-coordinates to use based on arclengths
        if a_xy_norm > 0:
            s_tot = 1/2 * (
                z_ins * np.sqrt( 1 + 4 * (a_xy_norm ** 2) * (z_ins ** 2) ) 
                + np.arcsinh(2 * a_xy_norm * z_ins) / ( 2 * a_xy_norm )
            )

        # if
        else:  # straight needle
            s_tot = z_ins

        # else

        s = np.unique( s[ s <= s_tot ] )  # remove longer and duplicate and sort

        dz = np.min( np.abs( np.diff( s ) ) )
        z = np.linspace( 0, z_ins, int( z_ins // dz + 1 ) )

        # compute 3D points
        points_z = np.hstack((
            a_xy.reshape( 1, -1 ) * z.reshape( -1, 1 ) ** 2, 
            z.reshape( -1, 1 )
        ))

        # interpolate to get s
        points, _ = numerical.interpolate_curve_s( points_z, s, axis=0 )

        return points

    # shape_quadratic


# class: AirDeflection

class ConstantCurvature:
    @staticmethod
    def k0( s: np.ndarray, kc: float, return_callable: bool = False ):
        """
            Intrinsic curvatures of the constant curvature needle

            :param s: numpy array of the arclengths
            :param kc: intrinsic curvature constant
            :param return_callable: (Default = False) returns the callable function

            :returns: (k0(s), k0'(s))
        """
        if return_callable:
            k0 = lambda s: kc
            k0prime = lambda s: 0

        # if
        else:
            k0 = kc * np.ones_like( s )
            k0prime = np.zeros_like( k0 )

        # else

        return k0, k0prime

    # k0

    @staticmethod
    def w0( s: np.ndarray, kc: float, thetaz: float = 0, return_callable: bool = False ):
        """
            Intrinsic curvatures of the constant curvature needle

            :param s: numpy array of the arclengths
            :param kc: intrinsic curvature constant
            :param thetaz: (Default = 0) the angle of rotation in the xy plane
            :param return_callable: (Default = False) returns the callable function

            :returns: (k0(s), k0'(s))
        """
        k0, k0prime = ConstantCurvature.k0( s, kc, return_callable=return_callable )

        # rotate the curvature
        Rz = geometry.rotz( thetaz )
        v = np.array( [ 1, 0, 0 ] ) @ Rz.T

        if return_callable:
            w0 = lambda s: k0( s ) * v
            w0prime = lambda s: k0prime( s ) * v

        # if
        else:
            w0 = k0 * v
            w0prime = k0prime * v

        # else

        return w0, w0prime

    # w0

    @staticmethod
    def shape( s: np.ndarray, curvature: Union[ float, np.ndarray ], thetaz: float = 0 ):
        """ Determine the (3D) constant curvature shape"""
        if isinstance( curvature, (int, float) ):
            curvature, _ = ConstantCurvature.w0( s, curvature, thetaz=thetaz )

        # if
        else:
            curvature = curvature.reshape( -1, 3 ).repeat( s.size, axis=0 ) @ geometry.rotz( thetaz ).T

        # else

        pmat, Rmat = numerical.integratePose_wv( curvature, s=s )

        return pmat, Rmat

    # shape


# class: ConstantCurvature

class SingleBend:
    @staticmethod
    def k0_1layer( s: np.ndarray, kc: float, length: float, return_callable: bool = False ):
        """
        Intrinsics curvatures of the double layer insertion scenario
        
        Original Author (MATLAB): Jin Seob Kim
        Translated Author: Dimitri Lezcano
        
        :param s: numpy array of the arclengths
        :param kc: intrinsic curvature constant
        :param length: length of needle insertion
        :param return_callable: (Default = False) returns the callable function

        :returns: (k0(s), k0'(s)) numpy arrays of shape s.shape
        """
        if return_callable:
            k0 = lambda s: kc * (1 - s / length) ** 2
            k0prime = lambda s: -2 * kc / length * (1 - s / length)

        # if
        else:
            k0 = kc * (1 - s / length) ** 2
            k0prime = -2 * kc / length * (1 - s / length)

        # else
        return k0, k0prime

    # k0_1layer

    @staticmethod
    def k0_2layer(
        s: np.ndarray, 
        kc_1: float, 
        kc_2: float, 
        length: float, 
        s_crit: float,
        return_callable: bool = False,
    ):
        """
        Intrinsics curvatures of the double layer insertion scenario
        
        Original Author (MATLAB): Jin Seob Kim
        Translated Author: Dimitri Lezcano
        
        :param s: numpy array of the arclengths
        :param kc_1: intrinsic curvature constant for layer 1
        :param kc_2: intrinsic curvature constant for layer 2
        :param length: length of needle insertion
        :param s_crit: arclength where needle shape boundary is
        :param return_callable: (Default = False) returns the callable function
        
        :returns: (k0(s), k0'(s)) numpy arrays of shape s.shape
        """
        if return_callable:
            def k0( s ):
                if s <= s_crit:
                    k0_s = kc_1 * ((s_crit - s) / length) ** 2 + kc_2 * (1 - s_crit / length) * (
                            1 + s_crit / length - 2 * s / length)

                # if
                else:
                    k0_s = kc_2 * (1 - s / length) ** 2

                # else

                return k0_s

            # k0

            def k0prime( s ):
                if s <= s_crit:
                    k0prime_s = -2 * kc_1 / length ** 2 * (s_crit - s) - 2 * kc_2 / length * (1 - s_crit / length)

                # if
                else:
                    k0prime_s = -2 * kc_2 / length * (1 - s / length)

                # else

                return k0prime_s

            # k0prime

        # if
        else:
            # set-up
            k0 = np.zeros_like( s )
            k0prime = np.zeros_like( s )

            # first layer processing
            s_1 = s[ s <= s_crit ]
            k0_1 = kc_1 * ((s_crit - s_1) / length) ** 2 + kc_2 * (1 - s_crit / length) * (
                    1 + s_crit / length - 2 * s_1 / length)
            k0prime_1 = -2 * kc_1 / length ** 2 * (s_crit - s_1) - 2 * kc_2 / length * (1 - s_crit / length)

            # second layer processing
            s_2 = s[ s > s_crit ]
            k0_2 = kc_2 * (1 - s_2 / length) ** 2
            k0prime_2 = -2 * kc_2 / length * (1 - s_2 / length)

            # set the return values
            k0[ s <= s_crit ] = k0_1
            k0[ s > s_crit ] = k0_2

            k0prime[ s <= s_crit ] = k0prime_1
            k0prime[ s > s_crit ] = k0prime_2

        # else

        return k0, k0prime

    # k0_2layer

    @staticmethod
    def k0_3layer(
        s: np.ndarray, 
        kc_1: float, 
        kc_2: float, 
        kc_3: float, 
        length: float, 
        s_crit_1: float,
        s_crit_2: float, 
        return_callable: bool = False,
    ):
        """
        Intrinsics curvatures of the double layer insertion scenario

        Original Author (MATLAB): Jin Seob Kim
        Translated Author: Dimitri Lezcano

        :param s: numpy array of the arclengths
        :param kc_1: intrinsic curvature constant for layer 1
        :param kc_2: intrinsic curvature constant for layer 2
        :param kc_3: intrinsic curvature constant for layer 3
        :param length: length of needle insertion
        :param s_crit_1: arclength where first needle shape boundary is
        :param s_crit_2: arclength where second needle shape boundary is
        :param return_callable: (Default = False) returns the callable function

        :returns: (k0(s), k0'(s)) numpy arrays of shape s.shape
        """
        if return_callable:
            def k0( s ):
                if s <= s_crit_1:
                    k0_s = kc_1 * (s_crit_1 - s) ** 2 / length ** 2 + kc_2 * (s_crit_2 - s_crit_1) * (
                            s_crit_2 + s_crit_1 - 2 * s) / length ** 2 + kc_3 * (1 - s_crit_2 / length) * (
                                   1 + (s_crit_2 - 2 * s) / length)

                # if
                elif s <= s_crit_2:
                    k0_s = kc_2 * (s_crit_2 - s) ** 2 / length ** 2 + kc_3 * (1 - s_crit_2 / length) * (
                            1 + s_crit_2 / length - 2 * s / length)

                # elif
                else:
                    k0_s = kc_3 * (1 - s / length) ** 2

                # else

                return k0_s

            # k0

            def k0prime( s ):
                if s <= s_crit_1:
                    k0prime_s = -2 * kc_1 / length ** 2 * (s_crit_1 - s) - 2 * kc_2 / length ** 2 * (
                            s_crit_2 - s_crit_1) - 2 * kc_3 / length * (1 - s_crit_2 / length)

                # if
                elif s <= s_crit_2:
                    k0prime_s = -2 * kc_2 / length ** 2 * (s_crit_2 - s) - 2 * kc_3 / length * (1 - s_crit_2 / length)

                # elif
                else:
                    k0prime_s = -2 * kc_3 / length * (1 - s / length)

                # else

                return k0prime_s

            # k0prime

        # if
        else:
            # set-up
            k0 = np.zeros_like( s )
            k0prime = np.zeros_like( s )

            # first layer processing
            mask_1 = s <= s_crit_1
            s_1 = s[ mask_1 ]
            # k0_1 = kc_1 * ((s_crit_1 - s_1) / length) ** 2 + kc_2 * (s_crit_2 - s_crit_1) / length * (
            #         s_crit_2 + s_crit_1)
            k0_1 = kc_1 * (s_crit_1 - s_1) ** 2 / length ** 2 + kc_2 * (s_crit_2 - s_crit_1) * (
                    s_crit_2 + s_crit_1 - 2 * s_1) / length ** 2 + kc_3 * (1 - s_crit_2 / length) * (
                           1 + (s_crit_2 - 2 * s_1) / length)

            k0prime_1 = -2 * kc_1 / length ** 2 * (s_crit_1 - s_1) - 2 * kc_2 / length ** 2 * (
                    s_crit_2 - s_crit_1) - 2 * kc_3 / length * (1 - s_crit_2 / length)

            # second layer processing
            mask_2 = (s_crit_1 < s) & (s <= s_crit_2)
            s_2 = s[ mask_2 ]
            k0_2 = kc_2 * (s_crit_2 - s_2) ** 2 / length ** 2 + kc_3 * (1 - s_crit_2 / length) * (
                    1 + s_crit_2 / length - 2 * s_2 / length)
            k0prime_2 = -2 * kc_2 / length ** 2 * (s_crit_2 - s_2) - 2 * kc_3 / length * (1 - s_crit_2 / length)

            # third layer processing
            mask_3 = s > s_crit_2
            s_3 = s[ mask_3 ]
            k0_3 = kc_3 * (1 - s_3 / length) ** 2
            k0prime_3 = -2 * kc_3 / length * (1 - s_3 / length)

            # set the return values
            k0[ mask_1 ] = k0_1
            k0[ mask_2 ] = k0_2
            k0[ mask_3 ] = k0_3

            k0prime[ mask_1 ] = k0prime_1
            k0prime[ mask_2 ] = k0prime_2
            k0prime[ mask_3 ] = k0prime_3

        # else

        return k0, k0prime

    # k0_3layer

    @staticmethod
    def determine_2layer_boundary(
        kc1: float, 
        length: float, 
        z_crit: float, 
        B: np.ndarray, 
        w_init: np.ndarray = None,
        s0: float = 0, 
        ds: float = 0.5, 
        R_init: np.ndarray = np.eye( 3 ),
        Binv: np.ndarray = None, 
        needle_rotations: list = None, 
        continuous: bool = False,
    ):
        """
            Determine the length of the needle that is inside the first layer only (s_crit)

            Original Author (MATLAB): Jin Seob Kim
            Translated Author: Dimitri Lezcano

            :param kc1: intrinsic curvature constant for layer 1
            :param s0: start position of s0
            :param length: length of needle insertion
            :param B: the needle stiffness matrix
            :param z_crit: length of the first layer
            :param w_init: (Default = [kc1; 0; 0]) the initial angular deformation
            :param s0: (Default = 0) a float of the inital length to integrate
            :param ds: (Default = 0.5) a float tof the ds to integrate
            :param R_init: (Default = 3x3 identity) SO3 matrix for initial rotation angle
            :param Binv: (Default = None) inv(B) Can be provided for numerical efficiency
            :param needle_rotations: (Default = None) list of needle axis rotations in radians
            :param continuous: (Default = False) whether to perform continuous integration

            :returns: s_crit: the critical arclength (rounded to the resolution of the arclength's ds)
                                (-1 if not in second-layer yet)
        """
        # w_init check
        if w_init is None:
            w_init = np.array( [ kc1, 0, 0 ] )

        # if

        s = np.arange( s0, length + ds, ds )

        # compute position of single-layer approximation
        if continuous:
            k0 = lambda s: SingleBend.k0_1layer( s, kc1, length )[ 0 ]
            k0prime = lambda s: SingleBend.k0_1layer( s, kc1, length )[ 1 ]

            w0 = lambda s: np.append( k0( s ), [ 0, 0 ] )
            w0prime = lambda s: np.append( k0prime( s ), [ 0, 0 ] )

            pmat_single, *_ = numerical.integrateEP_w0_ode(
                    w_init, w0, w0prime, B, s=s, R_init=R_init, Binv=Binv,
                    arg_check=False, needle_rotations=needle_rotations
            )
        # if
        else:
            k0, k0prime = SingleBend.k0_1layer( s, kc1, length )

            w0 = np.hstack( (k0.reshape( -1, 1 ), np.zeros( (k0.size, 2) )) )
            w0prime = np.hstack( (k0prime.reshape( -1, 1 ), np.zeros( (k0prime.size, 2) )) )

            pmat_single, *_ = numerical.integrateEP_w0(
                    w_init, w0, w0prime, B, s=s, R_init=R_init, Binv=Binv,
                    arg_check=False, needle_rotations=needle_rotations
            )

        # else

        # determine the point closest (but after) the boundary
        dz = pmat_single[ :, 2 ] - z_crit
        if np.all( dz <= 0 ):
            s_crit = -1  # not in double-layer yet

        # if
        else:
            s_crit_idx = np.argmin( np.abs( dz ) )
            s_crit = s[ s_crit_idx ]

        # else

        return s_crit

    # determine_2layer_boundary


# class: SingleBend

class DoubleBend:
    @staticmethod
    def k0_1layer(
        s: np.ndarray, 
        kc: float, 
        length: float, 
        s_crit: float, 
        p: float = 2 / 3,
        return_callable: bool = False,
    ):
        """
            Intrinsics curvatures of the double layer insertion scenario

            Original Author (MATLAB): Jin Seob Kim
            Translated Author: Dimitri Lezcano

            :param s: numpy array of the arclengths
            :param kc: intrinsic curvature constant
            :param length: length of needle insertion
            :param s_crit: the 180 degree turn insertion depth
            :param p: (Default = 2/3) float of the kappa_c scaling parameter
            :param return_callable: (Default = False) returns the callable function

            :returns: (k0(s), k0'(s)) numpy arrays of shape s.shape
        """
        if return_callable:
            # scale kappa_c values
            kc1 = kc * (s_crit / length) ** p
            kc2 = kc * (1 - s_crit / length) ** p

            def k0( s ):
                if s < s_crit:
                    k0_s = kc1 * (1 - s / length) ** 2

                elif s == s_crit:
                    k0_s = 1 / 2 * (kc1 - kc2) * (1 - s / length) ** 2

                else:
                    k0_s = -kc2 * (1 - s / length) ** 2

                return k0_s

            # k0

            def k0prime( s ):
                if s < s_crit:
                    k0prime_s = -2 * kc1 / length * (1 - s1 / length)

                elif s == s_crit:
                    k0prime_s = -(kc1 - kc2) * (1 - s / length)

                else:
                    k0prime_s = -2 * kc2 / length * (1 - s2 / length)

                return k0prime_s

            # k0prime
        # if
        else:
            # arclength setups (before & after double-bend)
            s1 = s[ s <= s_crit ]
            s2 = s[ s >= s_crit ]

            # kappa_c values
            kc1 = kc * ((s1.max() - s1.min()) / length) ** p
            kc2 = kc * ((s2.max() - s2.min()) / length) ** p

            # kappa_0 calculations
            k0_1 = kc1 * (1 - s1 / length) ** 2
            k0_2 = -kc2 * (1 - s2 / length) ** 2
            k0_12 = 1 / 2 * (k0_1[ -1 ] + k0_2[ 0 ])

            k0 = np.hstack( (k0_1[ :-1 ], k0_12, k0_2[ 1: ]) )

            # kappa_0' calculations
            k0prime_1 = -2 * kc1 / length * (1 - s1 / length)
            k0prime_2 = -2 * kc2 / length * (1 - s2 / length)
            k0prime_12 = 1 / 2 * (k0prime_1[ -1 ] + k0prime_2[ 0 ])

            k0prime = np.hstack( (k0prime_1[ :-1 ], k0prime_12, k0prime_2[ 1: ]) )

        # else

        return k0, k0prime

    # k0_1layer

# class: DoubleBend



"""

Library of needle shape sensing intrinic measurement functions

Author: Dimitri Lezcano

"""

from enum import Flag
from typing import Union, Callable

import tensorflow as tf

from . import numerical, geometry


class SHAPETYPE( Flag ):
    # first byte is for the number of bends, the second byte is for the number of layers ( - 1)
    CONSTANT_CURVATURE = 0x00
    SINGLEBEND_SINGLELAYER = 0x01
    SINGLEBEND_DOUBLELAYER = 0x02
    DOUBLEBEND_SINGLELAYER = 0x10 | SINGLEBEND_SINGLELAYER

    def get_k0( self ):
        """ Get the intrinsics kappa 0 function

        """
        k0_fns = {
                # SHAPETYPE.CONSTANT_CURVATURE    : ConstantCurvature.k0,
                SHAPETYPE.SINGLEBEND_SINGLELAYER: SingleBend.k0_1layer,
                SHAPETYPE.SINGLEBEND_DOUBLELAYER: SingleBend.k0_2layer,
                SHAPETYPE.DOUBLEBEND_SINGLELAYER: DoubleBend.k0_1layer,
        }

        return k0_fns[ self ]

    # get_k0

    def get_shape_class( self ):
        """ Get the intrinsics shape class

        """
        classes = {
                # SHAPETYPE.CONSTANT_CURVATURE    : ConstantCurvature,
                SHAPETYPE.SINGLEBEND_SINGLELAYER: SingleBend,
                SHAPETYPE.SINGLEBEND_DOUBLELAYER: SingleBend,
                SHAPETYPE.DOUBLEBEND_SINGLELAYER: DoubleBend,
        }

        return classes[ self ]

    # get_shape_class

    def k0( self, *args, **kwargs ):
        """ Get kappa_0 """
        return self.get_k0()( *args, **kwargs )

    # k0

    def w0( self, *args, **kwargs ):
        """ Get omega_o intrinsic 3D curvature function """
        k0 = self.k0( *args, **kwargs )
        if callable( k0 ):
            w0 = lambda s: tf.convert_to_tensor( [ k0( s ), 0, 0 ], dtype=k0( s ).dtype )

        # if
        else:
            w0 = tf.concat(
                    (k0, tf.zeros( (2, k0.shape[ 0 ]), dtype=k0.dtype )),
                    axis=0
            )

        # else

        return w0

    # w0


# enum class: SHAPETYPES

class SingleBend:
    @staticmethod
    def k0_1layer( s: tf.Tensor, kc: float, length: float, return_callable: bool = False ) -> (
            Union[ tf.Tensor, Callable ], Union[ tf.Tensor, Callable ]):
        """
        Intrinsics curvatures of the double layer insertion scenario

        Original Author (MATLAB): Jin Seob Kim
        Translated Author: Dimitri Lezcano

        :param s: numpy array of the arclengths
        :param kc: intrinsic curvature constant
        :param length: length of needle insertion
        :param return_callable: (Default = False) returns the callable function

        :returns: (k0(s), k0'(s)) numpy arrays of shape s.shape
        """
        if return_callable:
            k0 = lambda s: kc * (1 - s / length) ** 2
            k0prime = lambda s: -2 * kc / length * (1 - s / length)

        # if
        else:
            k0 = kc * (1 - s / length) ** 2
            k0prime = -2 * kc / length * (1 - s / length)

        # else
        return k0, k0prime

    # k0_1layer

    @staticmethod
    def k0_2layer(
            s: tf.Tensor, kc_1: float, kc_2: float, length: float, s_crit: float,
            return_callable: bool = False
    ):
        """
        Intrinsics curvatures of the double layer insertion scenario

        Original Author (MATLAB): Jin Seob Kim
        Translated Author: Dimitri Lezcano

        :param s: numpy array of the arclengths
        :param kc_1: intrinsic curvature constant for layer 1
        :param kc_2: intrinsic curvature constant for layer 2
        :param length: length of needle insertion
        :param s_crit: arclength where needle shape boundary is
        :param return_callable: (Default = False) returns the callable function

        :returns: (k0(s), k0'(s)) numpy arrays of shape s.shape
        """
        if return_callable:
            def k0( s ):
                if s <= s_crit:
                    k0_s = kc_1 * ((s_crit - s) / length) ** 2 + kc_2 * (1 - s_crit / length) * (
                            1 + s_crit / length - 2 * s / length)

                # if
                else:
                    k0_s = kc_2 * (1 - s / length) ** 2

                # else

                return k0_s

            # k0

            def k0prime( s ):
                if s <= s_crit:
                    k0prime_s = -2 * kc_1 / length ** 2 * (s_crit - s) - 2 * kc_2 / length * (1 - s_crit / length)

                # if
                else:
                    k0prime_s = -2 * kc_2 / length * (1 - s / length)

                # else

                return k0prime_s

            # k0prime

        # if
        else:
            # first layer processing
            s_1 = tf.boolean_mask( s, s <= s_crit )
            k0_1 = kc_1 * ((s_crit - s_1) / length) ** 2 + kc_2 * (1 - s_crit / length) * (
                    1 + s_crit / length - 2 * s_1 / length)
            k0prime_1 = -2 * kc_1 / length ** 2 * (s_crit - s_1) - 2 * kc_2 / length * (1 - s_crit / length)

            # second layer processing
            s_2 = tf.boolean_mask( s, s > s_crit )
            k0_2 = kc_2 * (1 - s_2 / length) ** 2
            k0prime_2 = -2 * kc_2 / length * (1 - s_2 / length)

            # set the return values
            k0 = tf.concat( (k0_1, k0_2), axis=0 )
            k0prime = tf.concat( (k0prime_1, k0prime_2), axis=0 )

        # else

        return k0, k0prime

    # k0_2layer

    @staticmethod
    def k0_3layer(
            s: tf.Tensor, kc_1: float, kc_2: float, kc_3: float, length: float, s_crit_1: float,
            s_crit_2: float, return_callable: bool = False
    ):
        """
        Intrinsics curvatures of the double layer insertion scenario

        Original Author (MATLAB): Jin Seob Kim
        Translated Author: Dimitri Lezcano

        :param s: numpy array of the arclengths
        :param kc_1: intrinsic curvature constant for layer 1
        :param kc_2: intrinsic curvature constant for layer 2
        :param kc_3: intrinsic curvature constant for layer 3
        :param length: length of needle insertion
        :param s_crit_1: arclength where first needle shape boundary is
        :param s_crit_2: arclength where second needle shape boundary is
        :param return_callable: (Default = False) returns the callable function

        :returns: (k0(s), k0'(s)) numpy arrays of shape s.shape
        """
        if return_callable:
            def k0( s ):
                if s <= s_crit_1:
                    k0_s = kc_1 * (s_crit_1 - s) ** 2 / length ** 2 + kc_2 * (s_crit_2 - s_crit_1) * (
                            s_crit_2 + s_crit_1 - 2 * s) / length ** 2 + kc_3 * (1 - s_crit_2 / length) * (
                                   1 + (s_crit_2 - 2 * s) / length)

                # if
                elif s <= s_crit_2:
                    k0_s = kc_2 * (s_crit_2 - s) ** 2 / length ** 2 + kc_3 * (1 - s_crit_2 / length) * (
                            1 + s_crit_2 / length - 2 * s / length)

                # elif
                else:
                    k0_s = kc_3 * (1 - s / length) ** 2

                # else

                return k0_s

            # k0

            def k0prime( s ):
                if s <= s_crit_1:
                    k0prime_s = -2 * kc_1 / length ** 2 * (s_crit_1 - s) - 2 * kc_2 / length ** 2 * (
                            s_crit_2 - s_crit_1) - 2 * kc_3 / length * (1 - s_crit_2 / length)

                # if
                elif s <= s_crit_2:
                    k0prime_s = -2 * kc_2 / length ** 2 * (s_crit_2 - s) - 2 * kc_3 / length * (1 - s_crit_2 / length)

                # elif
                else:
                    k0prime_s = -2 * kc_3 / length * (1 - s / length)

                # else

                return k0prime_s

            # k0prime

        # if
        else:

            # first layer processing
            s_1 = tf.boolean_mask( s, s <= s_crit_1 )
            # k0_1 = kc_1 * ((s_crit_1 - s_1) / length) ** 2 + kc_2 * (s_crit_2 - s_crit_1) / length * (
            #         s_crit_2 + s_crit_1)
            k0_1 = kc_1 * (s_crit_1 - s_1) ** 2 / length ** 2 + kc_2 * (s_crit_2 - s_crit_1) * (
                    s_crit_2 + s_crit_1 - 2 * s_1) / length ** 2 + kc_3 * (1 - s_crit_2 / length) * (
                           1 + (s_crit_2 - 2 * s_1) / length)

            k0prime_1 = -2 * kc_1 / length ** 2 * (s_crit_1 - s_1) - 2 * kc_2 / length ** 2 * (
                    s_crit_2 - s_crit_1) - 2 * kc_3 / length * (1 - s_crit_2 / length)

            # second layer processing
            s_2 = tf.boolean_mask( s, (s_crit_1 < s) & (s <= s_crit_2) )
            k0_2 = kc_2 * (s_crit_2 - s_2) ** 2 / length ** 2 + kc_3 * (1 - s_crit_2 / length) * (
                    1 + s_crit_2 / length - 2 * s_2 / length)
            k0prime_2 = -2 * kc_2 / length ** 2 * (s_crit_2 - s_2) - 2 * kc_3 / length * (1 - s_crit_2 / length)

            # third layer processing
            s_3 = tf.boolean_mask( s, s > s_crit_2 )
            k0_3 = kc_3 * (1 - s_3 / length) ** 2
            k0prime_3 = -2 * kc_3 / length * (1 - s_3 / length)

            # set the return values
            k0 = tf.concat( (k0_1, k0_2, k0_3), axis=0 )
            k0prime = tf.concat( (k0prime_1, k0prime_2, k0prime_3), axis=0 )

        # else

        return k0, k0prime

    # k0_3layer


# class: SingleBend

class DoubleBend:
    @staticmethod
    def k0_1layer(
            s: tf.Tensor, kc: float, length: float, s_crit: float, p: float = 2 / 3,
            return_callable: bool = False
    ):
        """
            Intrinsics curvatures of the double layer insertion scenario

            Original Author (MATLAB): Jin Seob Kim
            Translated Author: Dimitri Lezcano

            :param s: numpy array of the arclengths
            :param kc: intrinsic curvature constant
            :param length: length of needle insertion
            :param s_crit: the 180 degree turn insertion depth
            :param p: (Default = 2/3) float of the kappa_c scaling parameter
            :param return_callable: (Default = False) returns the callable function

            :returns: (k0(s), k0'(s)) numpy arrays of shape s.shape
        """
        if return_callable:
            # scale kappa_c values
            kc1 = kc * (s_crit / length) ** p
            kc2 = kc * (1 - s_crit / length) ** p

            def k0( s ):
                if s < s_crit:
                    k0_s = kc1 * (1 - s / length) ** 2

                elif s == s_crit:
                    k0_s = 1 / 2 * (kc1 - kc2) * (1 - s / length) ** 2

                else:
                    k0_s = -kc2 * (1 - s / length) ** 2

                return k0_s

            # k0

            def k0prime( s ):
                if s < s_crit:
                    k0prime_s = -2 * kc1 / length * (1 - s1 / length)

                elif s == s_crit:
                    k0prime_s = -(kc1 - kc2) * (1 - s / length)

                else:
                    k0prime_s = -2 * kc2 / length * (1 - s2 / length)

                return k0prime_s

            # k0prime
        # if
        else:
            # arclength setups (before & after double-bend)
            s1 = tf.boolean_mask( s, s <= s_crit )
            s2 = tf.boolean_mask( s, s >= s_crit )

            # kappa_c values
            kc1 = kc * ((tf.reduce_max( s1 ) - tf.reduce_min( s1 )) / length) ** p
            kc2 = kc * ((tf.reduce_max( s2 ) - tf.reduce_min( s2 )) / length) ** p

            # kappa_0 calculations
            k0_1 = kc1 * (1 - s1 / length) ** 2
            k0_2 = -kc2 * (1 - s2 / length) ** 2
            k0_12 = 1 / 2 * (k0_1[ -1 ] + k0_2[ 0 ])

            k0 = tf.concat( (k0_1[ :-1 ], [k0_12], k0_2[ 1: ]), axis=0 )

            # kappa_0' calculations
            k0prime_1 = -2 * kc1 / length * (1 - s1 / length)
            k0prime_2 = -2 * kc2 / length * (1 - s2 / length)
            k0prime_12 = 1 / 2 * (k0prime_1[ -1 ] + k0prime_2[ 0 ])

            k0prime = tf.concat( (k0prime_1[ :-1 ], [k0prime_12], k0prime_2[ 1: ]), axis=0 )

        # else

        return k0, k0prime

    # k0_1layer

# class: DoubleBend



"""

Library of rigid body mechanical geometry functions.

Author: Dimitri Lezcano

"""

from typing import Union

import numpy as np
from spatialmath import SO2, SO3, SE3


def arclength( points: np.ndarray, axis: int = 0 ):
    """ Compute the arclength of a curve

        :param points: An N x M numpy array  of curve points
        :param axis: the axis in which the points are given by

        :return: (total length, arclength for each point

    """
    # if (axis == 1):
    #     points = points.T # transpose the matrix
    #
    # # if

    # compute differences
    dpoints = np.diff( points, axis=axis )
    ds = np.linalg.norm( dpoints, axis=(axis + 1) % 2 )

    # compute lengths
    s = np.append( [ 0 ], np.cumsum( ds ) )

    return np.max( s ), s


# arclength

def axangle2rotm( theta: Union[ float, np.ndarray ], w: np.ndarray ) -> np.ndarray:
    """ Compute a rotation matrix from the axis angle representation

        :param theta: the angle of rotation (in radians)
        :param w: the unit vector of the direction of rotation

        :return: 3x3 Rotation matrix about w for theta rads
    """

    return rodrigues( theta * w )


# axangle2rotm


def hat( x: Union[ int, float, np.ndarray ] ) -> np.ndarray:
    """ Perform the inverse vee operation

        Args:
            x: a float or a numpy 1-d array of size [0, 1,3,6]

        Returns:
            2x2 so(2) element if x.size == 0
            3x3 so(3) element if x.size == 3
            4x4 se(3) element if x.size == 6
    """
    if isinstance( x, (int, float) ):
        X = hat_so2( x )

    else:
        assert (x.ndim == 1)

        if x.size == 1:  # so(2)
            X = hat_so2( x[ 0 ] )

        elif x.size == 3:  # so(3)
            X = hat_so3( x )

        elif x.size == 6:  # se(3)
            X = hat_se3( x )

        else:
            raise IndexError( "Size of `x` must be either 1, 3, or 6 vector" )

    # else

    return X


# hat


def hat_so2( x: float ) -> np.ndarray:
    """ Perform inverse vee operation for so(2)"""

    return np.array( [ [ 0, -x ], [ x, 0 ] ] )


# hat_so2

def hat_so3( x: np.ndarray ) -> np.ndarray:
    """ Perform inverse vee operation for so(3)"""
    assert (x.ndim == 1 and x.size == 3)

    X = np.zeros( (3, 3) )
    X[ 0, 1 ] = -x[ 2 ]
    X[ 0, 2 ] = x[ 1 ]
    X[ 1, 2 ] = -x[ 0 ]

    X -= X.T

    return X


# hat_so3

def hat_se3( x: np.ndarray ) -> np.ndarray:
    """ Perform inverse vee operation for se(3)
        Args:
            x = [angular; linear] = [w;v]

        Return:
            4x4 se(3) element array

    """
    assert (x.ndim == 1 and x.size == 6)

    X = np.zeros( (4, 4) )

    # perform rotational term
    X[ 0:3, 0:3 ] = hat_so3( x[ 0:3 ] )

    # perform linear term
    X[ 0:3, -1 ] = x[ 3:6 ]

    return X


# hat_se3

def is_so2( X: np.ndarray ) -> bool:
    """ Criteria for so2 matrix

        - 2x2 matrix
        - skew-symmetric

    """

    return X.shape == (2, 2) and is_skewsymm( X )


# is_so2

def is_SO2( R: np.ndarray ) -> bool:
    """ Determine if R is an element of SO(3)"""
    try:
        SO2( R, check=True )
        retval = True

    except ValueError:
        retval = False

    return retval


# is_SO2

def is_so3( X: np.ndarray ) -> bool:
    """ Criteria for so2 matrix

        - 3x3 matrix
        - skew-symmetric

    """

    return X.shape == (3, 3) and is_skewsymm( X )


# is_so3

def is_SO3( R: np.ndarray ) -> bool:
    """ Determine if R is an element of SO(3)"""
    try:
        SO3( R, check=True )
        retval = True

    except ValueError:
        retval = False

    return retval


# is_SO3


def is_se3( X: np.ndarray ) -> bool:
    """ Criteria for so2 matrix

        - 4x4 matrix
        - so3 upper left block
        - bottom row of zeros

    """

    return X.shape == (4, 4) and is_skewsymm( X[ 0:3, 0:3 ] ) and np.all( X[ -1, : ] == 0 )


# is_so3

def is_SE3( H: np.ndarray ) -> bool:
    """ Determine if R is an element of SO(3)"""
    try:
        SE3( H, check=True )
        retval = (H.shape == (4, 4))

    except ValueError:
        retval = False

    return retval


# is_SE3


def is_skewsymm( X: np.ndarray ) -> bool:
    """ Determine if a matrix is skew-symmetric"""

    return np.all( X.T == -X )


# is_skewsymm

def is_symm( X: np.ndarray ) -> bool:
    """ Determine if a matrix is symmetric """

    return np.all( X.T == X )


# is_symm

def quat2rotm( q: np.ndarray ) -> np.ndarray:
    """ Convert a unit quaternion to a """
    q = q.astype( float )
    if np.linalg.norm( q ) > 0:
        q /= np.linalg.norm( q )  # ensure unit
    else:
        q = np.array( [ 1, 0, 0, 0 ], dtype=np.float64 )

    # unpack quaternion
    c, v = q[ 0 ], q[ 1:4 ]  # cos(theta/2), sin(theta/2) * w
    s = np.linalg.norm( v )

    # compute angle and vector
    theta = 2 * np.arctan2( s, c )
    if theta == 0:
        R = np.eye( 3 )  # identity matrix

    else:
        w = v / s

        # compute rotation matrix
        R = axangle2rotm( theta, w )

    # else

    return R


# quat2rotm


def rodrigues( w: np.ndarray ) -> np.ndarray:
    """ Compute the rodrigues formula for the vector

        :param w: 3D vector of rotation

        :return: 3x3 Rotation matrix

    """
    theta = np.linalg.norm( w )
    if theta == 0:
        R = np.eye( 3 )

    else:
        wh = hat_so3( w / theta )  # skew(w)
        R = np.eye( 3 ) + np.sin( theta ) * wh + (1 - np.cos( theta )) * (wh @ wh)

    return R


# rodrigues


def rotm2axangle( R: np.ndarray ) -> (float, np.ndarray):
    """ Convert a rotation matrix to a quaternion
        :param R: numpy array of rotation matrix

        :return: angle of rotation, 3-D vector unit vector
    """
    theta = float( np.arccos( (np.trace( R ) - 1) / 2 ) )
    if theta == 0:
        w = np.array( [ 1, 0, 0 ] )

    else:
        w = 1 / (2 * np.sin( theta )) * vee_so3( R - R.T )

    return theta, w.astype( np.float64 )


# rotm2axangle

def rotm2quat( R: np.ndarray ) -> np.ndarray:
    """ Convert a rotation matrix to a quaternion
        :param R: numpy array of rotation matrix

        :return: 4-D vector unit quaternion of (w, x, y, z)
    """

    theta, w = rotm2axangle( R )
    w = w.astype( float )
    if np.linalg.norm( w ) > 0:
        w /= np.linalg.norm( w )

    return np.append( np.cos( theta / 2 ), np.sin( theta / 2 ) * w )


# rotm2quat

def rot2d( t: float ) -> np.ndarray:
    """ 2D rotation matrix"""

    return np.array( [ [ np.cos( t ), -np.sin( t ) ], [ np.sin( t ), np.cos( t ) ] ] )


# rot2d

def rotx( t: float ) -> np.ndarray:
    """ Rotation matrix about x-axis"""
    return np.array(
            [ [ 1, 0, 0 ], [ 0, np.cos( t ), -np.sin( t ) ], [ 0, np.sin( t ), np.cos( t ) ] ] )


# rotx

def roty( t: float ) -> np.ndarray:
    """ Rotation matrix about y-axis"""
    return np.array(
            [ [ np.cos( t ), 0, np.sin( t ) ], [ 0, 1, 0 ], [ -np.sin( t ), 0, np.cos( t ) ] ] )


# roty

def rotz( t: float ) -> np.ndarray:
    """ Rotation matrix about z-axis"""
    return np.array(
            [ [ np.cos( t ), -np.sin( t ), 0 ], [ np.sin( t ), np.cos( t ), 0 ], [ 0, 0, 1 ] ] )


# rotz

def vee( X: np.ndarray ) -> Union[ float, np.ndarray ]:
    """ Perform vee operation on X

        Args:
            X: numpy array of either so(2), so(3), se(3)

        Return:
            float if X is so(2)
            numpy vector if X is so(3) or se(3)

    """

    if is_so2( X ):
        x = vee_so2( X )

    elif is_so3( X ):
        x = vee_so3( X )

    elif is_se3( X ):
        x = vee_se3( X )

    else:
        raise TypeError( "'X' is not so(2), so(3) or se(3)" )

    return x


# vee

def vee_so2( X: np.ndarray, validate: bool = False ) -> float:
    """ Perform vee operation on X

        Args:
            X: 2x2 so(2) array
            validate: check whether X is so(2) or not

        Return:
            float of vee(X) X[1,0]

    """

    if validate:
        assert (is_so2( X ))

    return X[ 1, 0 ]


# vee_so2

def vee_so3( X: np.ndarray, validate: bool = False ) -> np.ndarray:
    """ Perform vee operation on X

        Args:
            X: 3x3 so(3) array
            validate: boolean on whether to check if X is so(3)

        Return:
            float of vee(X) [ X[ 2, 1 ], X[ 0, 2 ], X[ 1, 0 ] ]

    """

    if validate:
        assert (is_so3( X ))

    x = np.array( [ X[ 2, 1 ], X[ 0, 2 ], X[ 1, 0 ] ] )

    return x


# vee_so3

def vee_se3( X: np.ndarray, validate: bool = False ) -> np.ndarray:
    """ Perform vee operation on X

        Args:
            X: 4x4 se(3) array
            validate: boolean on whether to check if X is so(3)

        Return:
            float of vee(X) [ vee_so3(X[0:3,0:3]), x[0:3,-1]]

    """

    if validate:
        assert (is_se3( X ))

    x = np.append( vee_so3( X[ 0:3, 0:3 ], validate=validate ), X[ 0:3, -1 ] )

    return x

# vee_se3



"""
Created on Jul 28, 2021

@author: Dimitri Lezcano

@summary: This is a library for processing FBG signals

"""
import numpy as np


def process_signals( signals:np.ndarray, ref_wavelengths:np.ndarray ) -> np.ndarray:
    """ Remove reference wavelengths from the signals

    """

    return signals - ref_wavelengths.reshape(1,-1)


# process_signals

def temperature_compensation( signal_shifts, num_channels: int, num_active_areas: int ):
    """ Perform temperature compensation over the signal shits for FBG signals

        signal_shifts: numpy array of size (N, num_channels*num_active_areas)
        num_channels: int of the number of channels for the FBG needle
        num_active_areas: int of the number of active areas for the FBG needle
    """

    active_area_assignments = list( range( 1, num_active_areas + 1 ) ) * num_channels
    active_area_assignments = np.array( active_area_assignments )

    # iterate over the active areas removing temperature compensation
    signal_shifts_tempcomp = signal_shifts.copy()
    for aa_i in range( 1, num_active_areas + 1 ):
        aa_i_mask = active_area_assignments == aa_i

        signal_shifts_aa_i_mean = np.mean( signal_shifts[ :, aa_i_mask ], axis=1, keepdims=True )
        signal_shifts_tempcomp[ :, aa_i_mask ] -= signal_shifts_aa_i_mean  # remove mean value

    # aa_i

    return signal_shifts_tempcomp

# temperature_compensation



from setuptools import setup, find_packages

setup(
        name='needle-shape-sensing',
        version='1.1.0',
        author='Dimitri Lezcano',
        author_email='dlezcan1@jhu.edu',
        packages=find_packages(),
        url='http://pypi.python.org/pypi/needle-shape-sensing/',
        license='LICENSE.txt',
        description='Needle Shape Sensing library',
        long_description=open( 'README.md' ).read(),
        install_requires=[
                'numpy',
                'scipy',
                'spatialmath-python',
                'sympy',
                'numba',
                'tensorflow',
                'torch',
                'torchvision',
                'torchaudio',
                'needle-shape-sensing',
        ],
        classifiers=[
                'Operating System :: OS Independent',
                'License :: OSI Approved :: MIT License'
        ]
)
